"""
    SecTheory.ConformanceDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import sys
import os
if len(sys.argv) > 1 and sys.argv[1].find("demo") >= 0:
    this_dir = os.path.dirname(os.path.abspath( __file__ ))
    sys.path.append(this_dir + '/..')
    import molass_legacy.KekLib, SerialAnalyzer, DataStructure, Decomposer

import numpy as np
from time import time
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import RadioButtons, CheckButtons, Slider
from scipy.interpolate import UnivariateSpline
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.Peaks.ElutionModels import egh_pdf, emg
from SecTheory.LanJorgensonEgh import compute_tau, compute_sigma

Ti = -1000

class SecConfCurve:
    def __init__(self, t0, P, rp, m, size=70):
        rg = np.linspace(5, 80, size)
        rho = rg/rp
        rho[rho > 1] = 1
        tr = t0 + P*(1 - rho)**m
        self.spline = UnivariateSpline(np.flip(tr), np.flip(rg))

    def __call__(self, x):
        return self.spline(x)

class DisplaySlider(Slider):
    def __init__(self, *args, **kwargs):
        Slider.__init__(self, *args, **kwargs)

    def _update(self, event):
        pass

    def on_changed(self, func):
        pass

class ConformanceDemo(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.radio2 = None
        self.radio2_ax = None
        self.slider_axes = None
        self.fixed_index = 0
        self.updating = -1
        self.check_list = None
        self.last_change = time()
        Dialog.__init__(self, parent, "SEC Conformance Demo", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig = plt.figure(figsize=(18,7))
        self.fig = fig

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.start_interactive()

        self.popup_menu = None
        # self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def start_interactive(self):
        fig = self.fig
        fig.suptitle("SEC Conformance Demo with Extended Plate Theory Constraints", fontsize=24)

        gs = GridSpec(1,3)
        ax1 = fig.add_subplot(gs[0,0:2])
        ax1.set_title("Retention Time, Particle Size and Band Broadening", fontsize=16)
        ax1.set_xlabel("Time (Elution No.)", fontsize=16)
        ax1.set_ylabel("Concentration", fontsize=16)

        N = 10000
        mu = 200
        tr = mu - Ti
        sigma = tr/np.sqrt(N)
        tau = 0
        self.currect_params = [N, mu, sigma, tau]

        self.x = x = np.linspace(100, 700, 600)
        self.curve, = ax1.plot(x, egh_pdf(x, mu, sigma, tau), color="orange", label="Elution Curve")
        ax1.legend(loc="upper center", fontsize=16)

        t0 = 120
        ymin, ymax = ax1.get_ylim()
        ymax_ = 0.05
        ax1.set_ylim(ymin, ymax_)
        ax1.plot([t0, t0], [ymin, ymax_], ":", color="red")
        dy = (ymax_ - ymin)*0.02
        ax1.text(t0, dy, r"$t_0$", fontsize=16)

        axt = ax1.twinx()
        axt.grid(False)
        axt.set_ylabel("Rg")

        rp = 100
        P = 700
        self.secconf_curve  = SecConfCurve(t0, P, rp, 3)

        x_ = x[x > t0]
        y_ = self.secconf_curve(x_)

        axt.plot(x_, y_, ":", color="gray", label="SEC Conformance Curve")
        axt.legend(loc="upper right", fontsize=16)
        axt.set_ylim(0, 90)

        rg_ = self.get_rg()
        self.rg_text = axt.text(400, 60, r"$R_g=%.1f$" % rg_, fontsize=40, alpha=0.3, ha="center", va="center")

        sigma_ = self.get_sigma()
        self.sigma_text = axt.text(400, 45, r"$\sigma=%.1f$" % sigma_, fontsize=40, alpha=0.3, ha="center", va="center")

        self.recs = [
            ["N", 5000, 16000, N],
            [r"$\mu$", 150, 650, mu],
            [r"$\sigma$", 10, 20, sigma],
            [r"$\tau$", 0, 10, tau],
            ]

        self.reset_radio2()

        rax = fig.add_axes([0.7, 0.65, 0.07, 0.2])
        rax.set_title("Models", fontsize=14)
        # radio1 = RadioButtons(rax, ('Gaussian', 'EGH', 'EMG'))
        radio1 = RadioButtons(rax, ('Gaussian', 'EGH'))
        radio1.on_clicked(self.radio1_clicked)
        self.radio1 = radio1
        self.current_model = radio1.value_selected

        self.gau_updaters = [self.slider0_update, self.slider1_update, self.slider2_update, self.slider3_update]
        self.egh_updaters = [self.slider0_egh_update, self.slider1_egh_update, self.slider2_egh_update, self.slider3_egh_update]

        self.reset_sliders([0, 1, 2], active_index=0)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
        self.mpl_canvas.draw()

    def get_rg(self):
        tr = self.currect_params[1]
        return self.secconf_curve(tr)

    def get_sigma(self):
        sigma, tau = self.currect_params[2:4]
        return np.sqrt(sigma**2 + tau**2)

    def radio1_clicked(self, label):
        print(label)
        self.current_model = label
        if label == 'Gaussian':
            self.reset_radio2(tau=False)
            self.currect_params[3] = 0
            self.reset_sliders([0, 1, 2], active_index=0)
        else:
            self.reset_check()
            self.reset_sliders([0, 1, 2, 3], active_index=0)

        fig = self.fig
        fig.canvas.draw_idle()

    def radio2_clicked(self, label):
        i = self.radio2_labels.index(label)
        print(label, [i])
        self.fixed_index = i
        self.reset_sliders(self.slider_indeces, active_index=i)

    def reset_radio2(self, tau=False):
        if self.radio2 is not None:
            # self.radio2.remove()
            pass

        self.check_list = None
        fig = self.fig
        rax = fig.add_axes([0.8, 0.65, 0.07, 0.2])
        rax.set_title("Parameters", fontsize=16)
        labels =  ("N", r"$\mu$", r"$\sigma$")
        if tau:
            labels += (r"$\tau$",)
        radio2 = RadioButtons(rax, labels)
        radio2.on_clicked(self.radio2_clicked)
        self.radio2_labels = labels
        self.radio2_ax = rax
        self.radio2 = radio2

    def check_clicked(self, label):
        i = self.radio2_labels.index(label)
        print("before", label, [i], self.check_list)
        self.check_list[i] = not self.check_list[i]
        print("after", label, [i], self.check_list)
        self.reset_sliders(self.slider_indeces, active_index=i)

    def reset_check(self):
        if self.radio2 is not None:
            # self.radio2.remove()
            pass

        fig = self.fig
        rax = fig.add_axes([0.8, 0.65, 0.07, 0.2])
        rax.set_title("Parameters", fontsize=16)
        labels =  ("N", r"$\mu$", r"$\sigma$", r"$\tau$")
        self.check_list = [True, True, False, False]
        radio2 = CheckButtons(rax, labels, actives=self.check_list)
        radio2.on_clicked(self.check_clicked)
        self.radio2_labels = labels
        self.radio2_ax = rax
        self.radio2 = radio2

    def slider0_update(self, val):
        if self.updating > 0:
            return

        # print([0], val)
        _, mu, sigma, tau = self.currect_params
        N = val
        self.updating = 0
        self.currect_params[0] = N

        if self.fixed_index == 1:
            tr = mu - Ti
            sigma = tr/np.sqrt(N)
            self.currect_params[2] = sigma
            self.sliders[2].set_val(sigma)
        else:
            # self.fixed_index == 2:
            tr = sigma*np.sqrt(N)
            mu = tr + Ti
            self.currect_params[1] = mu
            self.sliders[1].set_val(mu)

        y = egh_pdf(self.x, mu, sigma, tau)
        self.curve.set_ydata(y)

        self.last_change = time()
        self.after(200, self.slider_update_cleaner)
        self.fig.canvas.draw_idle()

    def slider_update_cleaner(self):
        interval = time() - self.last_change
        if interval > 1:
            self.updating = -1
        else:
            self.after(200, self.slider_update_cleaner)

    def slider1_update(self, val):
        if self.updating >= 0 and self.updating != 1:
            return

        # print([1], val)
        N, _, sigma, tau = self.currect_params
        mu = val
        self.updating = 1
        self.currect_params[1] = mu
        if self.fixed_index == 0:
            tr = mu - Ti
            sigma = tr/np.sqrt(N)
            self.currect_params[2] = sigma
            self.sliders[2].set_val(sigma)
        else:
            # self.fixed_index == 2:
            tr = mu - Ti
            N = int((tr/sigma)**2)
            self.currect_params[0] = N
            self.sliders[0].set_val(N)

        y = egh_pdf(self.x, mu, sigma, tau)
        self.curve.set_ydata(y)
        self.update_texts()

        self.last_change = time()
        self.after(200, self.slider_update_cleaner)
        self.fig.canvas.draw_idle()

    def update_texts(self):
        rg_ = self.get_rg()
        self.rg_text.set_text(r"$R_g=%.1f$" % rg_)
        sigma_ = self.get_sigma()
        self.sigma_text.set_text(r"$\sigma=%.1f$" % sigma_)

    def slider2_update(self, val):
        if self.updating >= 0 and self.updating != 2:
            return

        # print([2], val)
        N, mu, _, tau = self.currect_params
        sigma = val
        self.updating = 2
        self.currect_params[2] = sigma

        if self.fixed_index == 0:
            tr = sigma*np.sqrt(N)
            mu = tr + Ti
            self.currect_params[1] = mu
            self.sliders[1].set_val(mu)
        else:
            # self.fixed_index == 1
            tr = mu - Ti
            N = int((tr/sigma)**2)
            self.currect_params[0] = N
            self.sliders[0].set_val(N)

        y = egh_pdf(self.x, mu, sigma, tau)
        self.curve.set_ydata(y)
        self.update_texts()

        self.last_change = time()
        self.after(200, self.slider_update_cleaner)
        self.fig.canvas.draw_idle()

    def slider3_update(self, val):
        print([3], val)
        self.fig.canvas.draw_idle()

    def slider0_egh_update(self, val):
        print([0], val)
        self.fig.canvas.draw_idle()

    def slider1_egh_update(self, val):
        print([1], val)
        self.fig.canvas.draw_idle()

    def slider2_egh_update(self, val):
        if self.updating >= 0 and self.updating != 2:
            return

        # print([2], val)
        N, mu, _, tau = self.currect_params
        sigma = val
        self.updating = 2
        self.currect_params[2] = sigma

        tau = compute_tau(Ti, N, mu, sigma)
        self.currect_params[3] = tau
        self.sliders[3].set_val(tau)

        y = egh_pdf(self.x, mu, sigma, tau)
        self.curve.set_ydata(y)
        self.update_texts()

        self.last_change = time()
        self.after(200, self.slider_update_cleaner)
        self.fig.canvas.draw_idle()

    def slider3_egh_update(self, val):
        if self.updating >= 0 and self.updating != 3:
            return

        # print([3], val)
        N, mu, sigma, _ = self.currect_params
        tau = val
        self.updating = 3
        self.currect_params[3] = tau

        sigma = compute_sigma(Ti, N, mu, tau)
        self.currect_params[2] = sigma
        self.sliders[2].set_val(sigma)

        y = egh_pdf(self.x, mu, sigma, tau)
        self.curve.set_ydata(y)
        self.update_texts()

        self.last_change = time()
        self.after(200, self.slider_update_cleaner)
        self.fig.canvas.draw_idle()

    def reset_sliders(self, indeces, active_index=0):
        if self.slider_axes is not None:
            for ax in self.slider_axes:
                ax.remove()

        for rec, val in zip(self.recs, self.currect_params):
            rec[-1] = val

        fig = self.fig
        slider_axes = []
        sliders = []
        for i, j in enumerate(indeces):
            label, valmin, valmax, valinit =  self.recs[j]
            ax = fig.add_axes([0.72, 0.55 - 0.1*i, 0.22, 0.03])
            slider_axes.append(ax)
            if self.check_list is None:
                display_only = j == active_index
            else:
                display_only = self.check_list[j]
            if display_only:
                handle_style = dict(size=0)
                slider  = DisplaySlider(ax, label=label, valmin=valmin, valmax=valmax, valinit=valinit, handle_style=handle_style)
            else:
                slider  = Slider(ax, label=label, valmin=valmin, valmax=valmax, valinit=valinit)

            if self.check_list is None:
                slider.on_changed(self.gau_updaters[j])
            else:
                slider.on_changed(self.egh_updaters[j])
            sliders.append(slider)
        self.slider_indeces = indeces
        self.slider_axes = slider_axes
        self.sliders = sliders

def demo():
    from molass_legacy.KekLib.TkUtils import get_tk_root
    import seaborn
    seaborn.set()
    root = get_tk_root()
    def show_dialog():
        dialog = ConformanceDemo(root)
        dialog.show()
        dialog.quit()
    root.after(0, show_dialog)
    root.mainloop()

if __name__ == '__main__':
    demo()



# coding: utf-8
"""
    ParamsAdjuster.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""

class ParamsAdjuster:
    def __init__(self, dsets, n_components, init_params):
        self.dsets = dsets
        self.n_components = n_components
        self.init_params = init_params

    def get_params_for(self, n):
        num_diff = n - self.self.n_components
        if num_diff == 0:
            ret_params = self.init_params
        elif num_diff > 0:
            ret_params = self.add_new_components(num_diff)
        else:
            ret_params = self.remove_extra_components(-num_diff)
        return ret_params

    def add_new_components(self, num):
        pass

    def remove_extra_components(self, num):
        pass



"""
    Distance.JensenShannon.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.spatial import distance
from molass_legacy.Trimming.Sigmoid import sigmoid, sigmoid_inv
import molass_legacy.KekLib.DebugPlot as plt

def deformed_jsd(y1, y2, b=None, debug=False):
    jsd = 0.5*distance.jensenshannon(y1, y2)
    a = np.percentile(y1, 95)
    if b is None:
        b = np.percentile(y2, 95)
    area_ratio = abs(a - b)/(a + b)
    ret_score = max(-100, 5*sigmoid_inv(max(jsd, area_ratio), 1, 0.01, 10.0, 0) + 0.5)

    if debug:
        print("jsd=", jsd, "area_ratio=", area_ratio, "ret_score=", ret_score)
        x = np.arange(len(y2))
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("stochastic_measure debug")
            ax.plot(x, y1)
            ax.plot(x, y2)
            fig.tight_layout()
            plt.show()

    return ret_score



# coding: utf-8
"""
    MatrixFactorAlternate.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from scipy.interpolate import UnivariateSpline
from molass_legacy.Models.ElutionCurveModels import emg
import molass_legacy.KekLib.DebugPlot as plt

def design_figure():
    fig = plt.figure()
    ax = fig.gca()
    x = np.arange(100)
    y1 = emg(x, h=1, mu=50, sigma=5)

    spline = UnivariateSpline([0, 20, 40, 60, 80, 99], [0.2, 0.11, 0.1, 0.03, 0.02, -0.1])
    # y2 = -0.003*x + 0.2
    y2 = spline(x)

    ax.set_title("Decomposition of Concentration", fontsize=16)
    ax.plot(x, y1, ':', linewidth=3, label='nomal elution element')
    ax.plot(x, y2, ':', linewidth=3, label='baseline drift')
    ax.plot(x, y1+y2, ':', color='red', linewidth=3, label='total elution')

    ax.legend(fontsize=16)
    fig.tight_layout()
    plt.show()

class MatrixFactorAlternate:
    def __init__(self, sd):
        self.q = sd.intensity_array[0,:,0]
        self.data = sd.intensity_array[:,:,1].T
        self.e_curve = sd.xray_curve

        peak_i = self.e_curve.primary_peak_i
        xslice = slice(0,100)
        yslice = slice(peak_i-50, peak_i+50)
        y = np.arange(self.data.shape[1])[yslice]

        x = self.q[xslice]
        xx, yy = np.meshgrid( x, y )
        i = np.arange(xslice.start, xslice.stop)
        j = np.arange(yslice.start, yslice.stop)
        ii, jj = np.meshgrid( i, j )

        zz = self.data[ii, jj]

        fig = plt.figure(figsize=(10, 8))
        ax = fig.add_subplot(111, projection='3d')
        ax.plot_surface(xx, yy, zz, color='green', alpha=0.3 )

        plt.show()




# coding: utf-8
"""
    LinearityScore.py

    Objects to facilitate garbage collection

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from scipy          import stats

DEBUG_PLOT  = False
if DEBUG_PLOT:
    import matplotlib.pyplot    as plt

FACTOR_WEIGHT       = 20

def linearity_score100( values ):
    return [ float( '%.3g' % ( np.power(abs(x),4)*FACTOR_WEIGHT) ) for x in values ]

def stderror_score100( values ):
    return [ float( '%.3g' % ( np.exp(-x*20)*FACTOR_WEIGHT ) ) for x in values ]

class LinearityScore:
    def __init__( self, x, y ): 
        base_angle = np.pi/4

        try:
            self.slope = slope = stats.linregress( x, y )[0]
            eval_angle = np.arctan( slope )

            # rotate to the base_angle so that slopes near zero won't have too low linearities
            angle = base_angle - eval_angle
            c = np.cos( angle )
            s = np.sin( angle )
            x_ = c*x - s*y
            y_ = s*x + c*y
            self.r_value, _, self.stderr = stats.linregress( x_, y_ )[2:5]

            if DEBUG_PLOT:
                print( slope, angle )
                plt.plot( x, y )
                plt.plot( x_, y_ )
                plt.show()
        except:
            import logging
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            logger = logging.getLogger( __name__ )
            logger.warning( 'caused an error in computing the linearity score: ' + etb.last_line() )
            self.slope      = np.nan
            self.r_value    = 0
            # self.stderr     = np.inf   # would result in an error in ExcelCOM.py
            self.stderr     = 1e10

    def get_params( self ):
        return self.slope, self.r_value, self.stderr



# basin hopping global optimization for the ackley multimodal objective function
from scipy.optimize import basinhopping
import numpy as np
from numpy.random import rand
from numpy import exp
from numpy import sqrt
from numpy import cos
from numpy import e
from numpy import pi
from numpy import meshgrid
from numpy import arange
import matplotlib.pyplot as plt
from matplotlib import animation

# objective function
def objective(v):
    x, y = v
    return -20.0 * exp(-0.2 * sqrt(0.5 * (x**2 + y**2))) - exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y))) + e + 20

# define range for input
r_min, r_max = -5.0, 5.0
# define the starting point as a random sample from the domain

minima_list = []
init_points = [(-4, -4), (-4, -4), (-3, 4), (-3, 4), (4, -3), (4, -3)]

def solve(k):
    # pt = r_min + rand(2) * (r_max - r_min)
    pt = init_points[k]
    fv = objective(pt)
    minima_list.append((pt, fv, True))

    if True:
        def minimum_callback(x, fv, accept):
            # print(x, fv, accept)
            minima_list.append((x, fv, accept))

        # perform the basin hopping search
        result = basinhopping(objective, pt, stepsize=0.5, niter=18, callback=minimum_callback)
    else:
        from scipy.optimize import differential_evolution
        def minimum_callback(xk, convergence=0):
            # print(x, fv, accept)
            fv = objective(xk)
            minima_list.append((xk, fv, convergence))

        # define range for input
        r_min, r_max = -5.0, 5.0
        # define the bounds on the search
        bounds = [[r_min, r_max], [r_min, r_max]]
        # perform the differential evolution search
        result = differential_evolution(objective, bounds, callback=minimum_callback)
    return result

for k in range(len(init_points)):
    result = solve(k)
    # summarize the result
    print('Status : %s' % result['message'])
    print('Total Evaluations: %d' % result['nfev'])
    # evaluate solution
    solution = result['x']
    evaluation = objective(solution)
    print('Solution: f(%s) = %.5f' % (solution, evaluation))

# sample input range uniformly at 0.1 increments
xaxis = arange(r_min, r_max, 0.1)
yaxis = arange(r_min, r_max, 0.1)
# create a mesh from the axis
xx, yy = meshgrid(xaxis, yaxis)
# compute targets
results = objective((xx, yy)) 
# create a surface plot with the jet color scheme
fig = plt.figure(figsize=(16,8))
ax = fig.add_subplot(121, projection='3d')
ax2 = fig.add_subplot(122)

fig.suptitle("Demonstration of Basin Hopping Algorithm", fontsize=20)
ax.set_title("Surface Plot", fontsize=16)
ax2.set_title("Contour Plot", fontsize=16)
fig.tight_layout()

surface = ax.plot_surface(xx, yy, results, cmap='jet', alpha=0.3)
contuor = ax2.contourf(xx, yy, results, cmap='jet', alpha=0.3)
# show the plot

minima_array = np.array([np.concatenate([x, [v]]) for x, v, a in minima_list])
x, y, z = minima_array[0]
points, = ax.plot([x], [y], [z], 'o', color='cyan')
trace3, = ax.plot(*minima_array[0:1,:].T, color='red')
point, = ax2.plot(x, y, 'o', color='cyan')
trace2, = ax2.plot(*minima_array[0:1,0:2].T, color='red')

print("len(minima_list)=", len(minima_list))

def reset():
    x, y, z = minima_array[0]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    trace3.set_data(*minima_array[0:1,0:2].T)
    trace3.set_3d_properties(minima_array[0:1,2], 'z')
    point.set_data([x], [y])
    trace2.set_data(*minima_array[0:1, 0:2].T)
    return points, trace3, point, trace2

def animate(i):
    x, y, z = minima_array[i]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    point.set_data([x], [y])
    color = 'cyan' if i % 20 == 0 else 'red'
    points.set_color(color)
    point.set_color(color)
    j, r = divmod(i, 20)
    start = j*20
    stop = start + r + 1
    slice_ = slice(start, stop)
    trace3.set_data(*minima_array[slice_,0:2].T)
    trace3.set_3d_properties(minima_array[slice_,2], 'z')
    trace2.set_data(*minima_array[slice_,0:2].T)
    return points, trace3, point, trace2

anim = animation.FuncAnimation(fig, animate, frames=len(minima_list), blit=True, init_func=reset, interval=200)
# anim.save("anim.gif")

plt.show()



"""
    Baseline.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from .Convex import integrative_curve_convex

LPM_PERCENTILE = 10
NEG_RATIO_LIMIT = 0.3   # < 0.41 for 20200624_2 UV, > 0.16 20200624_2 Xray
END_WIDTH = 50

def integrative_curve(y, baseline=None, epsilon=1e-4, p_final=LPM_PERCENTILE, raise_=False, end_slices=None, return_ratio=False, debug=False):
    if end_slices is None:
        end_slices = (slice(0,END_WIDTH), slice(-END_WIDTH, None))
    assert len(end_slices) == 2

    zeros = np.zeros(len(y))
    if baseline is None:
        x = np.arange(len(y))
        sbl = ScatteringBaseline(y)
        A, B = sbl.solve()
        baseline = A*x+B

    # to avoid inappropriately resorting to integral in the wide angle range
    escape_ratio = p_final/100 + 0.1

    start_y = np.average(y[end_slices[0]])
    final_y = np.average(y[end_slices[1]])

    if debug:
        print("len(y)=", len(y), "end_slices=", end_slices)
        # print("start_y=", start_y, y[end_slices[0]])
        # print("final_y=", final_y, y[end_slices[1]])

    min_e = None
    min_baseline = None
    for k in range(5):
        yb  = y - baseline
        y_ = np.max([yb, zeros], axis=0)
        if raise_ and k > 0:
            neg_ratio = len(np.where(yb < 0)[0])/len(y)
            # print([k], 'neg_ratio=', neg_ratio)
            if neg_ratio < NEG_RATIO_LIMIT or neg_ratio < escape_ratio:
                pass
            else:
                if False:
                    plt.push()
                    fig, ax = plt.subplots()
                    ax.plot(y)
                    ax.plot(baseline)
                    fig.tight_layout()
                    plt.show()
                    plt.pop()
                assert False

        cy = np.cumsum(y_)
        height = cy[-1] - cy[0]
        base = start_y
        ratio = (final_y - start_y)/height
        new_baseline = base + cy*ratio
        e = np.max(np.abs(new_baseline - baseline))/abs(height)
        # print([k], 'e=', e)
        if e < epsilon:
            break

        if min_e is None or e < min_e:
            min_e = e
            min_baseline = new_baseline
        else:
            new_baseline = min_baseline
            break

        baseline = new_baseline

    if return_ratio:
        return new_baseline, ratio
    else:
        return new_baseline

def better_integrative_curve(y, baseline=None, p_final=LPM_PERCENTILE, end_slices=None):
    try:
        b = integrative_curve(y, baseline=baseline, p_final=p_final, raise_=True, end_slices=end_slices)
        convex = False
    except AssertionError:
        b = integrative_curve_convex(y)
        convex = True
    return b, convex

def compute_baseline(y, x=None, integral=False, end_slices=None, return_params=False, debug=False):
    if x is None:
        import logging
        from molass_legacy.KekLib.DebugUtils import show_call_stack
        show_call_stack("compute_baseline")
        logging.getLogger(__name__).warning("*************** compute_baseline without x is deprecated.")
        x = np.arange(len(y))

    sbl = ScatteringBaseline(y, x=x)
    A, B = sbl.solve(debug=debug)
    baseline = A*x+B

    if integral:
        baseline_, ratio = integrative_curve(y, baseline, end_slices=end_slices, return_ratio=True)
        if ratio > 0:
            baseline = baseline_
        else:
            # temporary fix due to a possibly buggy result
            ratio = 0
    else:
        ratio = 0

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print("integral=", integral, "ratio=", ratio)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("Baseline.compute_baseline")
            ax.plot(x, y)
            ax.plot(x, A*x + B, color="red")
            ax.plot(x, baseline, ":")
            fig.tight_layout()
            plt.show()

    if return_params:
        if integral:
            return baseline, (A, B, ratio)
        else:
            return baseline, (A, B)
    else:
        return baseline



# coding: utf-8
"""
    PlateauRegions.py

    平坦な領域の判別

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy    as np

NUM_POINTS_PLATEAU      = 10
NON_PLATEAU_VALUE       = -1
ACCEPTABLE_DIFF_RATE    = 0.2
ACCEPTABLE_RG_DIFF      = 3
SPLIT_TEST_GRAD_RATE    = 0.05
ACCEPTABLE_DEV_RATE     = 0.05
ACCEPTABLE_STD_RATIO    = 0.2
MINIMUM_HALF_POINTS     = 5

DEBUG   = False
continuous_rule10 = np.arange(10)

class PlateauRegions:
    def __init__( self, array, average=None ):
        self.array      = np.array( [ NON_PLATEAU_VALUE if x is None else x for x in array ] )
        self.gradient   = np.gradient( self.array )
        # print( 'self.gradient.shape=', self.gradient.shape )
        plateaus        = np.array( self.array  )
        if average is None:
            acceptable_diff = ACCEPTABLE_RG_DIFF
        else:
            acceptable_diff = average * ACCEPTABLE_DIFF_RATE

        print( 'acceptable_diff=', acceptable_diff )

        plateaus[ np.abs( self.gradient ) > acceptable_diff ] = NON_PLATEAU_VALUE

        size = len(self.array)
        for i in range( size ):
            ok = False
            for j in range( i-NUM_POINTS_PLATEAU+1, i+1 ):
                if j < 0 or j > size - NUM_POINTS_PLATEAU:
                    continue
                if ( plateaus[j:j+NUM_POINTS_PLATEAU] > 0 ).all():
                    ok = True
                    break
            if not ok:
                plateaus[j] = NON_PLATEAU_VALUE

        if DEBUG: print( 'np.where( plateaus > 0 )=',np.where(  plateaus > 0 )[0] )

        def check_region( start, stop ):
            checked_intervals = []
            if stop - start < NUM_POINTS_PLATEAU:
                plateaus[start:stop] = NON_PLATEAU_VALUE
                return checked_intervals

            region  = plateaus[start:stop]
            if DEBUG: print( (start, stop ), 'region=', region[0:5], region[-5:] )
            # pmin, p50, pmax = np.percentile( region, [ 0, 50, 100 ] )
            # print( 'pmin, p50, pmax=', pmin, p50, pmax )

            # check the compactness of primary values

            grad    = self.gradient[start:stop]
            if True: print( 'grad=', grad )
            w40 = start + np.where( grad < -SPLIT_TEST_GRAD_RATE*acceptable_diff )[0]
            w60 = start + np.where( grad > +SPLIT_TEST_GRAD_RATE*acceptable_diff )[0]
            # print( 'w40=', w40 )
            # print( 'w60=', w60 )
            split_candedates = []
            for wnn in [ w40, w60 ]:
                for j in range(0,len(wnn)-10):
                    wj = wnn[j]
                    # print( 'wnn[%d]=%d' % ( j, wj ) )
                    if ( wnn[j:j+10] == wj + continuous_rule10 ).all():
                        for k in range( j+11, len(wnn) ):
                            if not (  k <= len(wnn) and wnn[j:k] == wj + np.arange(k-j) ).all():
                                split_candedates.append( (wnn[j], wnn[k]) )
                                break
                        break

            if True: print( 'split_candedates=', split_candedates )
            if len( split_candedates ) == 1:
                f, t = split_candedates[0]
                if f - start >= NUM_POINTS_PLATEAU and stop - t >= NUM_POINTS_PLATEAU:
                    intervals = [ [ start, f ], [ t, stop ] ]
                    plateaus[f:t]   = NON_PLATEAU_VALUE
                else:
                    intervals = [ [ start, stop ] ]
            else:
                intervals = [ [ start, stop ] ]

            if True: print( 'intervals=', intervals )
            for start_, stop_ in intervals:
                if len( intervals ) == 2:
                    region_ = plateaus[start_:stop_]
                elif len( intervals ) == 1:
                    region_ = region
                else:
                    assert( False )

                if DEBUG: print( (start_, stop_ ), 'region_=',  region_[0:5], region_[-5:] )
                # remove inappropriate points at both ends
                p50 = np.percentile( region_, 50 )
                if DEBUG: print( 'p50=', p50 )
                # print( 'region_ - p50=', region_ - p50 )
                in_appropriate =  np.where( np.abs( region_ - p50 ) > p50 * ACCEPTABLE_DEV_RATE )[0]
                if True: print( 'in_appropriate=', in_appropriate )

                exclude_rule_lower = np.arange( len( in_appropriate ) )
                region_size = len(region_)
                exclude_rule_upper = np.arange( region_size-len( in_appropriate ), region_size )
                interval = [ start_, stop_ ]
                for i, rule in enumerate( [ exclude_rule_lower, exclude_rule_upper ] ):
                    if DEBUG: print( 'rule=',  rule )
                    index = np.where( in_appropriate == rule )[0]
                    if True: print( 'index=',  in_appropriate[index] )
                    if len( index ) > 0:
                        plateaus[ start_ + in_appropriate[index] ] = NON_PLATEAU_VALUE
                        if i == 0:
                            interval[i] = start_ + in_appropriate[ index[-1] ] + 1
                        else:
                            interval[i] = start_ + in_appropriate[ index[0] ]
                    else:
                        # ok. no change
                        pass

                f, t = interval
                if t - f < NUM_POINTS_PLATEAU:
                    ok = False
                else:
                    std = np.std( plateaus[f:t] )
                    std_ratio = std/p50
                    print( (f, t), 'std=', std, 'std_ratio=', std_ratio )
                    if std_ratio <= ACCEPTABLE_STD_RATIO:
                        ok = True
                    else:
                        ok = False

                if ok:
                    checked_intervals.append( interval )
                else:
                    plateaus[f:t] = NON_PLATEAU_VALUE

            print( 'checked_intervals=', checked_intervals )
            return checked_intervals

        def get_regions( y ):
            r = []
            start = None
            for i in range( size ):
                if start is None:
                    if y[i] >= 0:
                        start = i
                else:
                    if y[i] < 0:
                        checked_intervals = check_region( start, i )
                        for interval in checked_intervals:
                            r.append( interval )
                        start = None

            if start is not None:
                checked_intervals = check_region( start, size )
                for interval in checked_intervals:
                    r.append( interval )

            return r

        regions = get_regions( plateaus )

        self.plateaus   = plateaus
        self.regions    = regions

    def is_consistent( self, other ):
        for f, t  in other.regions:
            included = False
            m = ( f + t ) // 2
            for ff, tt in self.regions:
                if ff <= m and m < tt:
                    included = True
                    break
            if not included:
                return False

        return True

    def adjust( self, interval, positive_curve=None ):
        lower, middle, upper = interval
        #  middle  = ( lower + upper ) // 2
        print( 'adjust regions=', self.regions, 'middle=', middle )

        for region in self.regions:
            lb, ub = region
            if upper <= lb or ub <= lower:
            # if lb < middle or middle > ub:
                continue
            else:
                lower = min( lb,        middle - MINIMUM_HALF_POINTS )
                upper = max( ub - 1,    middle + MINIMUM_HALF_POINTS )
                if positive_curve is None:
                    return ( lower, upper )
                else:
                    # print( 'len(self.plateaus)=', len(self.plateaus) )
                    # print( 'len(positive_curve)=', len(positive_curve) )
                    assert( len(positive_curve) == len(self.plateaus) )

                    where_non_positive = np.where( positive_curve[lower:middle] <= 0 )[0]
                    # print( 'lower where_non_positive', where_non_positive )
                    lower_ = None
                    adjusted = False
                    if len( where_non_positive ) == 0:
                        lower_ = lower
                    else:
                        lower_ = lower + where_non_positive[-1] + 1
                        adjusted = True
                        # TODO: where positive_curve[middle] <= 0

                    where_non_positive = np.where( positive_curve[middle:upper] <= 0 )[0]
                    # print( 'upper where_non_positive', where_non_positive )
                    upper_ = None
                    if len( where_non_positive ) == 0:
                        upper_ = upper
                    else:
                        # TODO: untested
                        upper_ = middle + where_non_positive[0] - 1
                        adjusted = True
                        # TODO: where positive_curve[middle-1] <= 0

                    if adjusted:
                        print( 'adjusted by positive_curve: ', ( lower, upper ), '=>', ( lower_, upper_ ) )
                    return ( lower_, upper_ )

        return None



"""
    SecTools.CorMap.CormapMakerDialog.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import os
import logging
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Rectangle
from matplotlib.gridspec import GridSpec
from matplotlib.widgets import SpanSelector
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from SerialDataUtils import load_intensity_files
from SaferSpinbox import SaferSpinbox
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.Elution.CurveUtils import simple_plot
from DataUtils import get_in_folder
from molass_legacy.KekLib.TkUtils import split_geometry
from .CormapVersion import get_version_string
from .AngularUnit import angstrom_qv

ALLOW_SPAN_CHANGE = False
DATA_VIESW_NAMES = ["2D View", "3D View"]
CORREL_TYPE_NAMES = ["numpy corrcoef", "ATSAS DATCMP"]
TRANSPOSE_NAMES = ["Frames", "Angular"]

class CormapMakerDialog(Dialog):
    def __init__(self, parent, demo=False, elution=False, atsas=None):
        self.logger = logging.getLogger(__name__)
        self.demo = demo
        self.elution = elution
        if atsas is None:
            from Env.EnvInfo import get_global_env_info
            env_info = get_global_env_info()
            atsas = env_info.atsas_is_available
        self.threed = False
        self.atsas = atsas
        self.rect = None
        self.colorbar = None
        self.cbar_ax = None
        self.cormap_drawn = False
        self.showing_transposed = True
        self.showing_datcmp = False
        self.datafiles = None
        self.popup_menu1 = None
        self.popup_menu3 = None
        self.datcmp_data = None
        self.datcmp_data_input = None
        Dialog.__init__(self, parent, get_version_string(), visible=False)

    def body(self, body_frame):
        frame = Tk.Frame(body_frame)
        frame.pack(padx=20, pady=10)

        input_frame = Tk.Frame(frame)
        input_frame.pack(pady=20)

        label = Tk.Label(input_frame, text="Input Folder: ")
        label.pack(side=Tk.LEFT)

        self.in_folder = Tk.StringVar()
        self.fe = FolderEntry(input_frame, textvariable=self.in_folder, width=80,
                                on_entry_cb=self.on_in_folder_entry)
        self.fe.pack(side=Tk.LEFT)

        cframe = Tk.Frame(frame)
        cframe.pack(padx=20)

        self.fig = fig = plt.figure(figsize=(18,6))
        self.gs = gs = GridSpec(2,12)
        ax1 = fig.add_subplot(gs[:,0:4])
        ax2 = fig.add_subplot(gs[0,4:8])
        ax3 = fig.add_subplot(gs[1,4:8])
        ax4 = fig.add_subplot(gs[:,8:12])
        self.axes = [ax1, ax2, ax3, ax4]    # this must be a list to be able to repalce ax1
        self.reset_cbar_ax()

        self.draw_titles()

        # self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.85, left=0.05, right=0.94, bottom=0.1, wspace=1.8)

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        tframe = Tk.Frame(frame)
        tframe.pack(padx=0, fill=Tk.X)      # padx=0: suppress changes from toolbar cursor display
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT, fill=Tk.X)
        tframe_right = Tk.Frame(tframe)
        tframe_right.pack(side=Tk.RIGHT)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_left)
        self.toolbar.update()

        self.build_control_bar(tframe_right)

        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def reset_cbar_ax(self):
        if self.cbar_ax is not None:
            self.cbar_ax.remove()
        self.cbar_ax = self.fig.add_axes([0.89, 0.0, 0.06, 1.0])
        self.cbar_ax.set_axis_off()

    def build_control_bar(self, frame):
        span_state = Tk.NORMAL if ALLOW_SPAN_CHANGE else Tk.DISABLED
        self.span_selector = Tk.BooleanVar()
        w = Tk.Checkbutton(frame, text="Span Selector", variable=self.span_selector, state=span_state)
        w.pack(side=Tk.LEFT, padx=10)
        self.span_selector_cb = w
        self.span_selector.trace("w", self.span_selector_tracer)

        self.whole_data_view = Tk.StringVar()
        w = ttk.Combobox(frame, textvariable=self.whole_data_view, values=DATA_VIESW_NAMES, width=10)
        w.pack(side=Tk.LEFT, padx=10)
        w.current(0)
        self.whole_data_view_cbox = w
        self.whole_data_view.trace("w", self.whole_data_view_tracer)

        select_frame = Tk.Frame(frame)
        select_frame.pack(side=Tk.LEFT)

        self.build_select_frame(select_frame, span_state)

        space = Tk.Frame(frame, width=10)
        space.pack(side=Tk.LEFT)

        self.correl_type = Tk.StringVar()
        state = Tk.NORMAL if self.demo or self.atsas else Tk.DISABLED
        w = ttk.Combobox(frame, textvariable=self.correl_type, values=CORREL_TYPE_NAMES, width=14, state=state)
        w.pack(side=Tk.LEFT, padx=10)
        w.current(int(self.atsas))
        self.correl_type_cbox = w
        self.correl_type.trace("w", self.correl_type_tracer)

        self.transpose = Tk.StringVar()
        w = ttk.Combobox(frame, textvariable=self.transpose, values=TRANSPOSE_NAMES, width=10)
        w.pack(side=Tk.LEFT, padx=10)
        w.current(0)
        self.transpose.trace("w", self.transpose_tracer)

        button_frame = Tk.Frame(frame)
        button_frame.pack(side=Tk.LEFT)

        self.cormap_btn_blink = BlinkingFrame(button_frame)
        self.cormap_btn_blink.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self.cormap_btn_blink, text="Draw Cormap", width=12, command=self.draw_cormap, state=Tk.DISABLED)
        w.pack()
        self.cormap_btn_blink.objects = [w]
        self.cormap_btn = w

        w = Tk.Button(button_frame, text="Save", width=6, command=self.save, state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        self.save_btn = w

    def build_select_frame(self, frame, span_state):
        self.select_vars = []
        self.sb_widgets = []
        range_name = "Elution-Range:" if self.elution else "Frame-Range"
        label = Tk.Label(frame, text=range_name)
        label.pack(side=Tk.LEFT)
        for k, sb_namel in enumerate(["start", "end"]):
            label = Tk.Label(frame, text=sb_namel)
            label.pack(side=Tk.LEFT, padx=5)

            var = Tk.IntVar()
            self.select_vars.append(var)
            sb = SaferSpinbox(frame, textvariable=var,
                        from_=0, to=200, increment=1, 
                        justify=Tk.CENTER, width=6, state=span_state)
            sb.pack(side=Tk.LEFT, padx=5)
            sb.set_tracer(self.redraw_elution_rect)
            var.trace("w", sb.tracer)
            self.sb_widgets.append(sb)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        if self.demo:
            self.after(500, self.perpare_temp_demo)
        self._show()

    def perpare_temp_demo(self):
        from molass_legacy._MOLASS.SerialSettings import get_setting
        in_folder = get_setting("in_folder")
        self.in_folder.set(in_folder)
        self.on_in_folder_entry()

    def on_in_folder_entry(self):
        self.update()
        in_folder = self.in_folder.get()
        try:
            data_array, datafiles = load_intensity_files(in_folder)
            self.datcmp_data_input = None
            self.datafiles = datafiles
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "on_in_folder_entry: ")
            return

        in_folder = get_in_folder(in_folder.replace("\\", "/"))
        self.fig.suptitle("Correlation Analysis on %s" % in_folder, fontsize=20)
        for ax in self.axes:
            ax.cla()

        if self.colorbar is not None:
            self.colorbar.remove()
            self.colorbar = None
        self.draw_titles()
        self.draw_ranges(data_array)
        self.activate_buttons()

    def activate_buttons(self):
        self.cormap_btn.config(state=Tk.NORMAL)
        if self.atsas:
            self.correl_type_cbox.config(state=Tk.NORMAL)

    def draw_titles(self):
        ax1, ax2, ax3, ax4 = self.axes
        ax2.set_title("Scattering Range", fontsize=16)
        ax4.set_title("Correlation Map", fontsize=16)

    def draw_ranges(self, data_array):
        self.qv = angstrom_qv(data_array[0,:,0])
        self.D = data_array[:,:,1].T
        self.E = data_array[:,:,2].T

        ax1, ax2, ax3, ax4 = self.axes
        self.draw_sequence_range(ax1)
        self.draw_scattering_range(ax2, ax3)
        self.update_select_vars()

        self.mpl_canvas.draw()

    def draw_sequence_range(self, ax, update_vars=True):
        title1 = "Elution Range" if self.elution else "Frame Sequence Range"
        ax.set_title(title1, fontsize=16)

        seq_name = "seqno." if self.elution else "Frame No."
        ax.set_xlabel(seq_name)
        ax.set_ylabel("Intensity")

        picking_q = 0.02
        i = bisect_right(self.qv, picking_q)
        ey = self.D[i,:]
        ex = np.arange(len(ey))
        N = 11
        sey = np.convolve(ey, np.ones(N)/N, mode='same')

        if self.elution:
            ecurve = ElutionCurve(sey)
            self.ecurve = ecurve
            simple_plot(ax, ecurve, legend=False)
            paired_ranges = ecurve.get_default_paired_ranges()
            ranges = paired_ranges[0].get_fromto_list()
            f, t = ranges[0][0], ranges[-1][1]
            info = ecurve.get_primary_peak_info()
            j_pair = (info[0], info[2])
        else:
            ax.plot(ex, sey, label="intensities at q=%g" % picking_q)
            f, t = 0, len(ey)-1
            j_pair = tuple(int(round(v)) for  v in np.linspace(0, self.D.shape[1], 4)[[1,2]])

        self.seq_slice = slice(f, t+1)
        self.j_pair = j_pair
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        self.rect = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch(self.rect)

        self.vlines = []
        label = None
        for j in j_pair:
            line,  = ax.plot([j, j], [ymin, ymax], ':', color="yellow", label=label)
            self.vlines.append(line)
            label = "selected frames"

        ax.legend()

        if update_vars:
            jmax = len(ex) - 1
            for k, j in [(0,f), (1,t)]:
                self.select_vars[k].set(j)
                sb = self.sb_widgets[k]
                sb.config(to=jmax)

    def redraw_elution_rect(self, *args):
        if self.rect is None:
            return

        self.cormap_drawn = False
        self.cormap_btn_blink.start()
        ax = self.axes[0]
        ymin, ymax = ax.get_ylim()
        f, t = [w.get() for w in self.select_vars[0:2]]
        self.rect.set_xy((f, ymin))
        self.rect.set_width(t - f)
        self.seq_slice = slice(f, t+1)
        self.mpl_canvas.draw()

    def on_span_select(self, xmin, xmax):
        if self.rect is None:
            return

        f = max(0, int(round(xmin)))
        t = min(self.D.shape[1]-1, int(round(xmax)))
        for k, j in enumerate([f, t]):
            self.select_vars[k].set(j)

    def cormap_available(self):
        correl_type = self.correl_type.get()
        transpose = self.transpose.get()
        ret_ok = correl_type == CORREL_TYPE_NAMES[0] or transpose == TRANSPOSE_NAMES[0]
        if not ret_ok:
            self.show_not_implemented_error(correl_type, TRANSPOSE_NAMES[1])
        return ret_ok

    def show_not_implemented_error(self, correl_type, transpose):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        MessageBox.showerror("Not Implemented Error",
            '"%s" is not available with "%s"' % (transpose, correl_type),
            parent=self)

    def correl_type_tracer(self, *args):
        if not self.cormap_available():
            return

        self.cormap_drawn = False
        self.cormap_btn_blink.start()

    def transpose_tracer(self, *args):
        if self.rect is None:
            return

        if not self.cormap_available():
            return

        self.cormap_drawn = False
        self.cormap_btn_blink.start()

    def draw_scattering_range(self, ax2, ax3):
        ax2.set_yscale("log")
        ax2.set_ylabel("Intensity")

        # ax2.axes.get_xaxis().set_ticks([])
        ax2.set_xticklabels([])
        ax3.set_yscale("log")
        ax3.set_ylabel("Stdev")
        ax3.set_xlabel(r"$q(\AA^{-1})$")

        for j in self.j_pair:
            sy = self.D[:,j]
            ax2.plot(self.qv, sy, label="curve at frame %d" % j)
            ey = self.E[:,j]
            ax3.plot(self.qv, ey, label="curve at frame %d" % j)

        ax2.legend()
        ax3.legend()

    def update_select_vars(self):
        imax = self.D.shape[1]-1
        for k, i in [(0,0), (1,imax)]:
            self.select_vars[k].set(i)
            sb = self.sb_widgets[k]
            sb.config(to=imax)

    def draw_cormap(self):
        from .CormapMaker import CormapMaker

        transpose = self.transpose.get() == TRANSPOSE_NAMES[0]
        self.showing_transposed = transpose
        correl_type = self.correl_type.get()

        if correl_type == CORREL_TYPE_NAMES[1]:
            if not transpose:
                self.show_not_implemented_error(correl_type, TRANSPOSE_NAMES[1])
                return

        self.cormap_btn_blink.stop()

        if correl_type == CORREL_TYPE_NAMES[0]:
            M = self.D[:,self.seq_slice]
            datcmp_data = None
            from_datcmp_str = ""
            self.showing_datcmp = False
        else:
            M, datcmp_data = self.get_data_from_datcmp()
            from_datcmp_str = " (DATCMP)"
            self.showing_datcmp = True
        self.datcmp_data = datcmp_data
        seqno = np.arange(self.seq_slice.start, self.seq_slice.stop)
        seqname = "Eno." if self.elution else "Frame No."
        cm = CormapMaker(M, self.qv, seqno, seqname, transpose=transpose, from_datcmp=datcmp_data is not None)
        ax = self.axes[-1]
        ax.cla()
        self.reset_cbar_ax()
        title = "Correlation Map%s" % from_datcmp_str
        ax.set_title(title, fontsize=16)
        if self.colorbar is not None:
            self.colorbar.remove()
        self.colorbar = cm.draw(ax, self.cbar_ax)
        if self.showing_transposed:
            self.draw_selected(ax)
        self.mpl_canvas.draw()
        self.cormap_drawn = True

    def draw_selected(self, ax):
        self.selected_points = []
        for pair in [self.j_pair, reversed(self.j_pair)]:
            point, = ax.plot(*pair, 'o', color="yellow")
            self.selected_points.append(point)

    def on_figure_click(self, event):
        if not self.showing_transposed:
            return

        if event.button == 1:
            self.draw_pair_position(event)
        elif event.button == 3:
            if event.inaxes == self.axes[0]:
                self.show_popup_menu1(event)
            elif event.inaxes == self.axes[3] and self.showing_datcmp:
                self.show_popup_menu3(event)
            return

    def draw_pair_position(self, event):
        ax4 = self.axes[3]
        if not self.cormap_drawn or event.inaxes != ax4:
            return

        min_i = self.seq_slice.start
        max_i = self.seq_slice.stop - 1
        def get_index(x):
            return max(min_i, min(max_i, int(round(x))))

        self.j_pair = tuple(sorted([get_index(v) for v in [event.xdata, event.ydata]]))

        ax1, ax2, ax3 = self.axes[0:3]
        if self.threed:
            for k, line in zip(self.j_pair, self.pairframe_lines):
                ydata = np.ones(len(self.qv)) * k
                line.set_xdata(self.qv)
                line.set_ydata(ydata)
                # https://stackoverflow.com/questions/46685326/how-to-set-zdata-for-a-line3d-object-in-python-matplotlib
                line.set_3d_properties(self.D[:,k])
        else:
            ymin, ymax = ax1.get_ylim()
            for k, line in zip(self.j_pair, self.vlines):
                line.set_data([k, k], [ymin, ymax])

        for ax in [ax2, ax3]:
            ax.cla()

        ax2.set_title("Scattering Range", fontsize=16)
        self.draw_scattering_range(ax2, ax3)

        for k, pair in enumerate([self.j_pair, reversed(self.j_pair)]):
            self.selected_points[k].set_data(*pair)

        self.mpl_canvas.draw()

    def show_popup_menu1(self, event):
        self.create_popup_menu1(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu1.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu1(self, event):
        if self.popup_menu1 is None:
            self.popup_menu1 = Tk.Menu(self, tearoff=0 )
            if ALLOW_SPAN_CHANGE:
                self.popup_menu1.add_checkbutton(label="Span Selector", variable=self.span_selector)
            self.popup_menu1.add_command(label='Chage View (2D ⇔ 3D)', command=self.change_view_2d3d)

    def show_popup_menu3(self, event):
        self.create_popup_menu3(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu3.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu3(self, event):
        if self.popup_menu3 is None:
            self.popup_menu3 = Tk.Menu(self, tearoff=0 )
            self.popup_menu3.add_command(label='Show Pairwise Map', command=self.show_pairwise_map)

    def get_data_from_datcmp(self):
        from molass_legacy.ATSAS.DatCmp import run_datcmp
        if self.datcmp_data_input is None:
            files = self.datafiles[self.seq_slice]
            self.datcmp_data_input = run_datcmp(files=files, return_dict=True)
        return self.datcmp_data_input

    def save(self):
        pass

    def whole_data_view_tracer(self, *args):
        ax1 = self.axes[0]
        if False:
            ax1.remove()    # this causes the following error
            """
            Traceback (most recent call last):
              File "C:\Program Files\Python39\lib\site-packages\matplotlib\cbook\__init__.py", line 270, in process
                func(*args, **kwargs)
              File "C:\Program Files\Python39\lib\site-packages\mpl_toolkits\mplot3d\axes3d.py", line 1223, in _button_release
                toolbar = getattr(self.figure.canvas, "toolbar")
            AttributeError: 'NoneType' object has no attribute 'canvas'
            """
        else:
            # work-around to avoid the above mentioned error, which can be causing memory leaks
            ax1.cla()
            ax1.set_axis_off()

        self.threed = self.whole_data_view.get().find("3") >= 0
        if self.threed:
            dim = "3d"
        else:
            dim = None
        ax1 = self.fig.add_subplot(self.gs[:,0:4], projection=dim)
        self.axes[0] = ax1
        self.mpl_canvas.draw()

        if self.threed:
            self.draw_data_in_3d(ax1)
            selector_state = Tk.DISABLED
        else:
            self.draw_sequence_range(ax1, update_vars=False)
            selector_state = Tk.NORMAL

        self.span_selector_cb.config(state=selector_state)
        self.mpl_canvas.draw()

    def draw_data_in_3d(self, ax):
        from MatrixData import simple_plot_3d
        ax.set_xlabel("$q(\AA^{-1})$")
        ax.set_ylabel("Frame No.")
        ax.set_zlabel("Intensity")

        ax.set_title("3D View", fontsize=16)
        simple_plot_3d(ax, self.D, x=self.qv)
        self.pairframe_lines = []
        x = self.qv
        for j in self.j_pair:
            y = np.ones(len(x)) * j
            z = self.D[:,j]
            line, = ax.plot(x, y, z, label="Frame No. %d" % j)
            self.pairframe_lines.append(line)
        ax.legend()

    def span_selector_tracer(self, *args):
        span_selector = self.span_selector.get()
        if span_selector:
            self.span = SpanSelector(self.axes[0], self.on_span_select, 'horizontal', useblit=True,
                        props=dict(alpha=0.5))
        else:
            self.span = None

    def show_pairwise_map(self):
        from .PairwiseMap import PairwiseMap
        M = self.D[:,self.seq_slice]
        dialog = PairwiseMap(self, M, self.j_pair, self.qv, self.datcmp_data)
        dialog.show()

    def change_view_2d3d(self):
        view_name = self.whole_data_view.get()
        current = 0 if view_name == DATA_VIESW_NAMES[0] else 1
        next = 1 - current
        self.whole_data_view.set(DATA_VIESW_NAMES[next])
        self.whole_data_view_cbox.current(next)



# coding: utf-8
"""
    ArgParser.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

class ArgParser:
    def __init__(self, argv):
        self.argv = argv

    def get_dict(self):
        ret = {}
        for v in self.argv[1:]:
            k, p = v.split('=')
            key = k.replace('--', '')
            val = int(p) if p.isnumeric() else p
            ret[key] = val

        return ret



"""
    ValidComponents.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

VALID_COMPONENTS_THRESHOLD = 0.01

class ValidComponents:
    def __init__(self, nc):
        self.nc = nc

    def update(self, optimizer):
        self.vc_vector = optimizer.minima_props > VALID_COMPONENTS_THRESHOLD

    def get_valid_vector(self):
        return np.ones(self.nc, dtype=bool)


# coding: utf-8
"""
    SingleGmm.py.

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
USE_SKLEARN_GMM = False


class SingleGmm:
    def __init__(self, k, max_iter=1000):
        if USE_SKLEARN_GMM:
            from sklearn.mixture import GaussianMixture
            self.gmm = GaussianMixture(k, max_iter=max_iter, covariance_type='spherical')
        else:
            from .FoleyGmm import GMM
            self.gmm = GMM(k, 40)

    def fit(self, X=None):
        """
        exected to get the following attributes
            weights_
            means_
            covariances_
        """

        gmm = self.gmm

        if USE_SKLEARN_GMM:
            gmm.fit(X=X)
            print('means_=', gmm.means_)
            print('covariances_=', gmm.covariances_)
            print('weights_=', gmm.weights_)

            self.weights_ = gmm.weights_
            self.means_ = gmm.means_.flatten()
            self.covariances_ = gmm.covariances_
            self.gamma = None   # ?
        else:
            gmm.fit(X)
            print('mu=', gmm.mu)
            print('sigma=', gmm.sigma)
            print('pi=', gmm.pi)

            self.weights_ = gmm.pi.flatten()
            self.means_ = gmm.mu.flatten()
            self.covariances_ = gmm.sigma.flatten()
            self.gamma = gmm.gamma



"""
    BaselineGuinier.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""

import logging
import numpy as np
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, INTEGRAL_BASELINE, restore_default_setting

BASELINE_TYPE_NAMES = ["no correction", "linear", None, None, None, "integral"]
BASELINE_TYPE_CODES = [0, 1, 5]

class GuinierResult:
    def __init__(self, sg):
        try:
            self.guinier_start = sg.guinier_start
        except:
            pass
        try:
            self.guinier_stop = sg.guinier_stop
        except:
            pass
        try:
            self.basic_quality = sg.basic_quality
        except:
            pass

class BaselineGuinier:
    def __init__(self, D, E, q, i_index, e_slice, end_slices=None):
        from molass_legacy.DataStructure.LPM import LPM_3d      # moved due to ImportError: cannot import name 'LPM_3d' from partially initialized module 'LPM' (most likely due to a circular import) 

        self.logger = logging.getLogger(__name__)

        z = D[i_index,:]
        yp = np.average(D[:,e_slice], axis=1)
        ype = np.average(E[:,e_slice], axis=1)

        lpm1 = LPM_3d(D, ecurve_y=z, integral=False, for_all_q=True, e_index=i_index)
        yp_lpm1 = np.average(lpm1.data[:,e_slice], axis=1)

        lpm2 = LPM_3d(D, ecurve_y=z, integral=True, for_all_q=False, e_index=i_index, end_slices=end_slices)
        yp_lpm2 = np.average(lpm2.data[:,e_slice], axis=1)

        curve0 = np.array([q, yp, ype]).T
        curve1 = np.array([q, yp_lpm1, ype]).T
        curve2 = np.array([q, yp_lpm2, ype]).T
        sg0 = SimpleGuinier(curve0)
        sg1 = SimpleGuinier(curve1)
        sg2 = SimpleGuinier(curve2)

        try:
            guinier_length0 = q[sg0.guinier_stop] - q[sg0.guinier_start]
        except:
            guinier_length0 = 0
        try:
            guinier_length1 = q[sg1.guinier_stop] - q[sg1.guinier_start]
        except:
            guinier_length1 = 0
        try:
            guinier_length2 = q[sg2.guinier_stop] - q[sg2.guinier_start]
        except:
            guinier_length2 = 0
        self.logger.info("lengths of guinier regions are %.3g, %.3g, %.3g", guinier_length0, guinier_length1, guinier_length2)

        self.curves = [curve0, curve1, curve2]
        self.results = [GuinierResult(sg) for sg in [sg0, sg1, sg2]]    # return proxies
        self.lengths = [guinier_length0, guinier_length1, guinier_length2]

    def get_scattering_curves(self):
        return self.curves

    def get_guinier_results(self):
        return self.results

    def get_guinier_lengths(self):
        return self.lengths

    def get_xray_baseline_type(self):
        return BASELINE_TYPE_CODES[np.argmax(self.lengths)]

    def get_xray_baseline_type_second(self):
        # https://stackoverflow.com/questions/33181350/quickest-way-to-find-the-nth-largest-value-in-a-numpy-matrix/43171216
        return BASELINE_TYPE_CODES[np.argpartition(self.lengths, -2)[-2]]

def evaluate_descending_side_upturn(sd, debug=False):
    D, E, qv, ecurve = sd.get_xr_data_separate_ly()
    small_angle = slice(0, sd.xray_index)
    peak_info = ecurve.peak_info
    slice_pair = [slice(0, peak_info[0][0]), slice(peak_info[-1][-1], None)]
    left_average = np.average(D[small_angle,slice_pair[0]])
    right_average = np.average(D[small_angle,slice_pair[1]])
    upratio = right_average/left_average

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.DataStructure.MatrixData import simple_plot_3d
        print("upratio=", upratio)
        x = ecurve.x
        y = ecurve.y
        plt.push()
        fig = plt.figure(figsize=(14,7))
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122)
        simple_plot_3d(ax1, D)
        ax2.plot(x, y)
        for slice_ in slice_pair:
            ax2.plot(x[slice_], y[slice_], color='yellow')
        fig.tight_layout()
        plt.show()
        plt.pop()

    return upratio

HW = 5
def get_peaktop_slice(topx):
    return slice(topx-HW, topx+HW)

def create_bg(sd, fc, logger, debug=False):
    D_, E_, qv, e_curve = sd.get_xr_data_separate_ly()

    fc_slice = slice(*fc.get_mapped_flow_changes())     # this slice can be, or better be, replaced with slice(None, None)
                                                        # because the effect in the Xray side is always ignorable.
    i_slice = slice(0, None)
    j_slice = fc_slice
    D = D_[i_slice,j_slice]
    E = E_[i_slice,j_slice]
    q = qv[i_slice]

    end_slices = get_setting('manual_end_slices')
    if end_slices is None:
        # make sure to be clear about j0 handling
        end_slices = e_curve.get_end_slices()

    i_index = sd.xray_index
    start = 0 if j_slice.start is None else j_slice.start
    topx = e_curve.peak_info[e_curve.primary_peak_no][1] - start

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.Elution.CurveUtils import simple_plot
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("xray_baseline_type debug")
            ax1.set_title("Mapped Flow Change")
            ax2.set_title("Preliminary Guinier Analysis Range")

            simple_plot(ax1, e_curve, color="orange", legend=False)
            ymin, ymax = ax1.get_ylim()
            ax1.set_ylim()
            for j in [j_slice.start, j_slice.stop]:
                label = None if j is None else "mapped flowchange"
                ax1.plot([j, j], [ymin, ymax], color="yellow", label=label)

            ax1.legend()

            ax2.set_yscale("log")
            y = np.average(D_[:,topx-5:topx+6], axis=1)
            ax2.plot(qv, y)

            fig.tight_layout()
            plt.show()

    bg = BaselineGuinier(D, E, q, i_index, get_peaktop_slice(topx), end_slices=end_slices)
    return bg

def change_baseline_correction_defaults(sd, fc, logger, debug=False):
    """
        note that sd is not yet trimmed.
    """

    integral = False
    if hasattr(sd, 'mtd_elution') and sd.mtd_elution is not None:
        return integral

    if fc.has_special():
        uv_baseline_type = 4
        set_setting('uv_baseline_type', uv_baseline_type)
        logger.info("uv_baseline_type=%d is preferable for special type of flow changes.", uv_baseline_type)
    else:
        uv_baseline_type = get_setting('uv_baseline_type')

    xray_baseline_type = get_setting('xray_baseline_type')

    if xray_baseline_type is None:
        bg = create_bg(sd, fc, logger, debug=debug)
        xray_baseline_type = bg.get_xray_baseline_type()
    else:
        if uv_baseline_type == 4 and xray_baseline_type == 5:
            # temporary fix for cases such as 20190309_1
            bg = create_bg(sd, fc, logger, debug=debug)

    if uv_baseline_type == 4 and xray_baseline_type == 5:
        # temporary fix for cases such as 20170304
        xray_baseline_type_second = bg.get_xray_baseline_type_second()
        logger.info("xray_baseline_type is changed from %d to %d for the special type.", xray_baseline_type, xray_baseline_type_second)
        xray_baseline_type = xray_baseline_type_second

    if xray_baseline_type == 0:
        set_setting('xray_baseline_opt', 0)
    else:
        if xray_baseline_type == 5:
            integral = True
        set_setting('xray_baseline_opt', 1)

    set_setting('xray_baseline_type', xray_baseline_type)

    logger.info("changed default baseline correction: %s (xray_baseline_type=%d)", BASELINE_TYPE_NAMES[xray_baseline_type], xray_baseline_type)

    return integral



"""
    Selective/IftOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data

def try_ift_optimization(button_frame, devel=True):
    if devel:
        from importlib import reload
        import IFT.IftOptimizeImpl
        reload(IFT.IftOptimizeImpl)
    from IFT.IftOptimizeImpl import ift_optimize   
    print("try_ift_optimization")
    editor = button_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    D, E, qv, ecurve = editor.sd.get_xr_data_separate_ly()
    num_components = len(params_array)
    M = get_denoised_data(D, rank=num_components)

    opt_params_array = ift_optimize(M, E, qv, fx, y, model, params_array, debug=True)




# -*- coding: utf-8 -*-
"""

    ファイル名：   CuntomizedPandasTable.py

    処理内容：

       Pilatus 画像データ情報の一覧

"""
from __future__ import division, print_function, unicode_literals

import sys
import os
# import copy

if sys.version_info > (3,):
    import tkinter as Tk
    import tkinter.ttk as ttk
else:
    import Tkinter as Tk
    import ttk
    import tkFont

import pandas as pd
from pandastable import Table, TableModel

label_bg_color = 'gray25'

class PandasTable( Tk.Frame ):
    def __init__( self, parent, columns=[ [ 'col0', 100 ], [ 'col1', 100 ] , [ 'col2', 100 ] ], width=None, rowheight=16, height=None ):
        Tk.Frame.__init__( self, parent )

        # self.columns = copy.deepcopy( columns )
        self.columns = []
        for col in columns:
            self.columns.append( col[0] )

        self.table = None

    def destroy( self ):
        # Don't use this method
        if self.table:
            # self.table.destroy()
            pass

        self.table = None
        Tk.Frame.destroy( self  )

    def dummy( self, event ):
        pass

    def selectAll( self ):
        self.table.selectAll()

    def selectRows( self, begin, end ):
        self.table.setSelectedCells( begin, end, 0, len( self.columns ) )
        self.table.drawMultipleCells()

    def selectionClear( self ):
        self.table.selectNone()

    def import_array( self, data ):

        if len(data) > 0:
            columns_ = self.columns
        else:
            columns_ = []

        df = pd.DataFrame( data, columns=columns_ )

        if self.table == None:
            self.table = Table( self, dataframe=df )

            self.table.rowheaderwidth=30    # does not work. why?
            self.table.thefont = ( 'Microsoft Sans Serif', 9 )
            # self.table.thefont = ( 'Arial', 10 )
            # self.table.thefont = ( 'Arial Unicode MS', 10 )
            
            # self.table.thefont = ( 'ＭＳ Ｐゴシック', 10 )

            self.table.show()

            self.table.bind( "<Button-3>", self.dummy )                     # to remove popup-menu binding
            self.table.tablecolheader.bind( "<Button-3>", self.dummy )      # to remove popup-menu binding
            self.table.rowheader.bind( "<Button-3>", self.dummy )           # to remove popup-menu binding
            self.table.rowindexheader.bind( "<Button-3>", self.dummy )      # to remove popup-menu binding

        else:

            """
                borrowed from Table.clearTable
            """
            model = TableModel( dataframe=df )
            self.table.updateModel( model )
            self.table.redraw()
            self.table.show()

    def bindActionMenu( self, event, popup, corner=True ):
        self.table.bind( event, popup )
        self.table.rowheader.bind( event, popup )
        if corner:
            self.table.rowindexheader.bind( event, popup ) 

    def selectedRows( self ):
        # print( 'selectedRows: ', self.table.multiplerowlist )
        return self.table.multiplerowlist

    def selectedCell( self ):
        row = self.table.getSelectedRow()
        col = self.table.getSelectedColumn()
        print( [row, col] )
        return [row, col]



# coding: utf-8
"""
    DecompDriver.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os, sys
from time import time, sleep
from multiprocessing import current_process
import logging
import numpy as np
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_savetxt

class UserCancel(Exception): pass

class DecompLoghandler(logging.Handler):
    def __init__(self, out_folder):
        logging.Handler.__init__(self)
        self.fh = open(out_folder + '/decomp.log', 'w')
        self.formatter  = logging.Formatter('%(asctime)s %(message)s', '%H:%M:%S')

    def emit( self, record ):
        if record.levelno >= logging.INFO:
            try:
                log_message = self.formatter.format(record) + '\n'
                self.fh.write(log_message)
                self.fh.flush()
            except:
                # this error can occur when canceled
                pass

    def __del__(self):
        self.fh.close()

def run_decomp(job_id, out_folder, job_info, table, use_gpu):
    try:
        from .DecompProcessUtils import run_decomp_impl

        logger  = logging.getLogger()
        started = time()

        if sys.stdout is None:
            sys.stdout = open(os.devnull, 'w')

        last_rec = table[job_id]

        process = current_process()
        pid = process.pid
        file_info = last_rec[0]
        submitted = last_rec[7]

        cancel_flag = table[job_id][11]
        if cancel_flag:
            raise UserCancel('User Cancel')

        # should not make the out_folder when canceled before actual execution.
        if not os.path.exists(out_folder):
            mkdirs_with_retry(out_folder)

        logger.addHandler(DecompLoghandler(out_folder))
        logger.setLevel(logging.INFO)

        os.chdir(out_folder)

        q = job_info.q
        a = job_info.a
        e = job_info.e
        dmax = job_info.dmax

        file = file_info[1]
        np_savetxt(file, np.array([q, a, e]).T)

        table[job_id] = [file_info, pid, -1, *last_rec[3:7], submitted, started, None, 0, cancel_flag]

        def progress_cb(step, chi2, rg, volume):
            duration = time() - started
            cancel_flag = table[job_id][11]
            if cancel_flag:
                raise UserCancel('User Cancel')
            table[job_id] = [file_info, pid, 0, step, chi2, rg, volume, submitted, started, None, duration, 0]

        ret = run_decomp_impl(q, a, e, dmax, file, job_info.infile_name,
                                progress_cb=progress_cb, steps=None,    # steps=1000, etc. for debug
                                use_gpu=use_gpu,
                                )
        state = -3
    except UserCancel as exc:
        last_rec = table[job_id]
        ret = exc
        state = -5
        logger.info('Execution canceled by user.')
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
        etb = ExceptionTracebacker()
        print(etb)
        logger.error(str(etb))
        ret = etb
        last_rec = table[job_id]
        state = -4

    finished = time()
    duration = finished - started
    last_rec = table[job_id]
    table[job_id] = [file_info, None, state, *last_rec[3:9], finished, duration, 0]
    print('run_decomp', job_id, 'finish')
    return ret



# coding: utf-8
"""
    EdBoundary.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import molass_legacy.KekLib.DebugPlot as plt

SAFE_START_RATIO = 0.9
POSITIVE_SMALL_VALUE = 1e-10
BOUNDARY_RATIO = 0.05

def guess_boundary_value(data, debug=False):
    v = np.array(sorted(data.flatten()))

    v[v < POSITIVE_SMALL_VALUE] = POSITIVE_SMALL_VALUE

    n = len(v)
    start = int(n*SAFE_START_RATIO)
    v_ = np.log10(v)
    gv = np.gradient(v_)
    gv_ = gv[start:]
    gv_max = gv_.max()
    gv_boundary = gv_max*BOUNDARY_RATIO
    w = np.where(gv_ > gv_boundary)[0]
    i = start + w[0]

    if debug:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
        ax1, ax2 = axes
        ax1.plot(v_)
        ax1.plot(i, v_[i], 'o', color='red')
        ax2.plot(gv)
        ax2.plot(i, gv[i], 'o', color='red')
        fig.tight_layout()
        plt.show()

    return v[i]



"""
    NumpyUtils.py

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF
"""

import re
import numpy as np

def get_valid_index( X ):
    return np.logical_and( np.logical_not( np.isnan( X ) ), np.logical_not( np.isinf( X ) ) )

def _save_names( fh, names, delimiter ):
    line = delimiter.join( names ) + '\n'
    fh.write( bytearray( line, 'cp932' ) )  # utf8 or cp932

def np_savetxt( file, np_array, mode="w", column_names=None, fmt='%.18e'):
    if file.find( '.csv' ) > 0:
        delimiter_ = ','
    elif file.find( '.xlsx' ) > 0:
        assert( mode == "w" )
        np_savexlsx( file, np_array, column_names=column_names )
        return

    else:
        delimiter_ = '\t'

    if mode == 'a':
        with open( file, "ab" ) as fh:
            if column_names is not None:
                _save_names( fh, column_names, delimiter_ )
            np.savetxt( fh, np_array, fmt=fmt, delimiter=delimiter_ )
    else:
        with open( file, "wb" ) as fh:
            if column_names is not None:
                _save_names( fh, column_names, delimiter_ )
            np.savetxt( fh, np_array, fmt=fmt, delimiter=delimiter_ )

def np_savetxt_with_comments(file, np_array, comments):
    with open( file, "wb" ) as fh:
        for line in comments:
            fh.write(bytearray( line, 'cp932' ))

    np_savetxt(file, np_array, mode='a')

def np_savexlsx(file, np_array, column_names=None):
    assert False, "Not implemented"

data_line_re = re.compile(r'^\s*\d+')

default_encoding = 'cp932'

def np_loadtxt_robust( fname, usecols=None, delimiter=None, encoding=None, retry=True ):
    global default_encoding

    encoding_ = default_encoding if encoding is None else encoding

    try:
        return np_loadtxt( fname, usecols, delimiter, encoding_, retry )
    except UnicodeDecodeError as exc:
        if encoding is None:
            try:
                default_encoding = 'utf-8'
                return np_loadtxt( fname, usecols, delimiter, default_encoding, retry )
            except Exception as exc:
                raise exc
        else:
            raise exc
    except Exception as exc:
        raise exc

def np_loadtxt( fname, usecols=None, delimiter=None, encoding='utf-8', retry=True ):
    global first_data_line

    if type( fname ) == str:
        fh = open( fname, 'r', encoding=encoding )
        is_csv =  fname.find( '.csv' ) > 0
    else:
        fh = fname
        is_csv = False

    comment_lines = []
    first_data_line = None

    def generator( fh ):
        global first_data_line

        for line in fh:
            if type( line ) == bytes:
                line = line.decode()

            if data_line_re.match( line ):
                if first_data_line is None:
                    first_data_line = line
                yield line
            else:
                comment_lines.append( line )

    if delimiter is None:
        delimiter = ',' if is_csv else None

    try:
        array = np.loadtxt( generator( fh ), delimiter=delimiter, usecols=usecols )
    except UnicodeDecodeError as e:
        # should be handled by the caller
        raise e
    except Exception as e:
        if retry and delimiter is None:
            delim_ = ',' if first_data_line.find( ',' ) > 0 else None
            array, comment_lines = np_loadtxt( fname, delimiter=delim_,
                usecols=usecols, encoding=encoding,
                retry=False     # to avoid infinite recursive calls
                )
        else:
            raise e

    fh.close()

    return array, comment_lines

def get_safe_approximate_max( array ):
    val95, val99 = np.percentile( array[ np.isfinite( array ) ], [ 95, 99 ] )
    return val99 if abs( ( val99 - val95 ) / val99 ) < 0.1 else val95

def simply_safe_sprintf( fmt, *args ):
    try:
        s = fmt % tuple( args )
    except:
        try:
            s = fmt % tuple( [ np.nan if a is None else a for a in args ] )
        except:
            s = ' '.join( [ fmt ] + [ str( a ) for a in args ] )
            if fmt[-1] == '\n':
                s += '\n'
    return s

def warped_weight_vector( size, m ):
    assert size - 1 > m 

    n = size - m

    dm = 1/(m-1)
    w_L = np.arange(m)*0.5*dm

    dn = 1/(n-1)
    w_R = 0.5 + 0.5*( dn + np.arange(0, n)*(1-dn)/(n-1) )

    return np.hstack( [ w_L, w_R ] )

def more_warped_weight_vector( size, m ):
    assert size - 1 > m 

    n = size - m

    hm = m//2
    hn = n//2
    hsize = size - hm - hn

    w_Z = np.zeros( hm )

    dm = 1/(hm-1)
    w_L = np.arange(hm)*0.5*dm

    dn = 1/(hn-1)
    w_R = 0.5 + 0.5*( dn + np.arange(0, hn)*(1-dn)/(hn-1) )

    w_1 = np.ones( hn )

    return np.hstack( [ w_Z, w_L, w_R, w_1 ] )

"""
    https://stackoverflow.com/questions/14313510/how-to-calculate-moving-average-using-numpy
"""
def moving_average(a, n=3, keepsize=False) :
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    ret_ = ret[n - 1:] / n
    if keepsize:
        assert n % 2 == 1
        # TODO: for n % 2 == 0
        hn  = n//2
        ret_ = np.hstack( [ [ret_[0]]*hn, ret_, [ret_[-1]]*hn  ] )

    return ret_

def arg_less_percentile(x, p):
    kth = int(len(x)*p/100)
    kpp = np.argpartition(x, kth)
    return int(np.average(kpp[0:kth+1]))

def arg_more_percentile(x, p):
    kth = int(len(x)*p/100)
    kpp = np.argpartition(x, kth)
    return int(np.average(kpp[kth:]))

def get_proportional_points(x1, x2, proportions):
    return np.array([x1*(1-w) + x2*w for w in proportions])



"""
    Peaks.MomentsUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

def compute_moments(x, y):
    W = np.sum(y)
    M1 = np.sum(x*y)/W              # raw moment
    M2 = np.sum(y*(x-M1)**2)/W      # central moment
    M3 = np.sum(y*(x-M1)**3)/W      # central moment
    return W, M1, M2, M3


"""
    Theory.TdGrant2022.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DENSS.DenssUtils import fit_data
from DataUtils import get_in_folder

def demo(root, sd, pno=None, debug=True, crysol_int_files=None, use_bounded_lrf=True, use_almerge=False, rank=2):
    from bisect import bisect_right
    from .SynthesizedLRF import synthesized_lrf_spike as synthesized_lrf
    from SvdDenoise import get_denoised_data
    from .Rg import compute_corrected_Rg, compute_Rg
    from .SolidSphere import get_boundary_params_simple

    assert crysol_int_files is not None

    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    if pno is None:
        pno = ecurve.primary_peak_no

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    eslice = slice(f,t)
    x = ecurve.x
    y = ecurve.y

    M0 = M[:,eslice]

    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])
    b1, b2, k = get_boundary_params_simple(Rg)
    print("Rg=%g, b1=%g" % (Rg, b1))

    c_ = y[eslice]

    if rank == 2:
        M2 = get_denoised_data(M0, rank=2)
        C2 = np.array([c_, c_**2])
        P2 = M2 @ np.linalg.pinv(C2)

        if use_bounded_lrf:
            from BoundedLRF.BoundedLrfSolver import BoundedLrfSolver
            solver = BoundedLrfSolver(qv, M2, E[:,eslice], C=C2, i=sd.xray_index)
            lrf_result = solver.solve()
            # P_, C__, Rg, R_, L_, hK, hL, bq_bounds_, coerced_bq_ = lrf_result
            P = lrf_result[0]
        else:
            P = synthesized_lrf(qv, M0, c_, M2, P2, boundary=b1, k=k)

    elif rank == 1:
        M1 = get_denoised_data(M0, rank=1)
        C1 = np.array([c_])
        P1 = M1 @ np.linalg.pinv(C1)
        P = P1
    else:
        assert False

    pt = M[:,p]
    spt = P[:,0]

    qc2, Ic2, Icerr2, D = fit_data(qv, spt, E[:,p])

    if use_almerge:
        pass
    else:
        def compute_almerge_like_values(b):
            i = bisect_right(qv, b)
            j = bisect_right(qc2, b)
            pt_ = pt*Ic2[j]/pt[i]
            ptc = np.hstack([Ic2[j-i:j], pt_[i:]])
            qc1, Ic1, Icerr1, _ = fit_data(qv, ptc, E[:,p], D=D)
            return pt_, qc1, Ic1

        pt_, qc1, Ic1 = compute_almerge_like_values(b1)

    ncols = len(crysol_int_files) + 2

    fig, axes = plt.subplots(ncols=ncols, figsize=(7*ncols,7))
    lrf_text = "Bounded LRF" if use_bounded_lrf else "SynLRF"
    in_folder = get_in_folder()
    crysol_int = crysol_int_files[0]
    crysol_int_name = crysol_int.split('\\')[-1]
    fig.suptitle("%s vs. ALMERGE on %s, %s (using IFT in DENSS)" % (lrf_text, in_folder, crysol_int_name), fontsize=20)

    ax0, ax1, ax2 = axes[0:3]

    ax0.set_title("Used Elution Range", fontsize=16)
    ax0.plot(x, y)
    ymin, ymax = ax0.get_ylim()
    p = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )
    ax0.add_patch(p)
    ax0.set_xlabel("Eno")
    ax0.set_ylabel("Intensity")

    ax1.set_title("Peak Top, LRF and TdGrant", fontsize=16)
    if len(axes) == 3:
        ax2.set_title("TdGrant and CRYSOL", fontsize=16)
        ax3 = None
    else:
        ax3 = axes[3]
        ax2.set_title("TdGrant and CRYSOL2", fontsize=16)
        ax3.set_title("TdGrant and CRYSOL3", fontsize=16)

    axes_ = axes[1:]

    for ax in axes_:
        ax.set_yscale('log')

    ax1.plot(qv, pt_, label='Peak Top Data')
    label = 'Bounded LRF' if use_bounded_lrf else 'Syn LRF'
    ax1.plot(qv, spt, label=label)

    ymin, ymax = ax1.get_ylim()

    for ax in axes_:
        ax.set_ylim(ymin, ymax)
        if Ic1 is not None:
            ax.plot(qc1, Ic1, label="ALMERGE like IFT", color='C2')
        label = 'Bounded LRF IFT' if use_bounded_lrf else 'Syn LRF IFT'
        ax.plot(qc2, Ic2, label=label, color='C3')

    if crysol_int_files is not None:
        from CrysolUtils import np_loadtxt_crysol
        for j, ax in enumerate(axes_[1:]):
            # crysol_int = crysol_int_files[j]
            # name = crysol_int.split('\\')[-1][0:4]
            name = crysol_int_name[0:4]
            data, _ = np_loadtxt_crysol(crysol_int)
            print('data.shape=', data.shape)
            if data.shape[1] == 5:
                crysol_ver = 2
                K = 2
            else:
                crysol_ver = 3
                K = 2
            qv_ = data[:,0]
            for k in range(1,K):
                y = data[:,k].copy()
                y *= Ic2[0]/y[0]
                if ax3 is None:
                    ax.plot(qv_, y, label='%s-CRYSOL' % (name), color='C%d' % (3+k))
                else:
                    # older style when there were tow versions of crysol, i.e., crysol and crysol_30
                    ax.plot(qv_, y, label='%s-CRYSOL%d-%d' % (name, crysol_ver, k), color='C%d' % (3+k))

    for ax in axes_:
        if rank == 2:
            ax.plot([b1, b1], [ymin, ymax], ':', color='gray', label='Rank boundary')
        ax.set_xlabel("Q")
        ax.set_ylabel("Intensity")
        ax.legend()

    if rank == 2 and use_bounded_lrf:
        for ax in axes_:
            axt = ax.twinx()
            axt.grid(False)
            for k, bound in enumerate(lrf_result[7]):   # bq_bounds_
                label = None if k > 0 else "B(q) bounds"
                axt.plot(qv, bound, ":", color="red", label=label)
            axt.legend(loc="lower left")

    fig.tight_layout()
    fig.subplots_adjust(top=0.88)
    plt.show()



# coding: utf-8
"""
    MfCorres.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import get_color
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from DraggablePoints import DraggablePoints
from .SvdPrecision import plot_text

def demo_func_impl(fig):
    q = np.linspace(0.01, 0.6, 600)
    rg_list = [50, 24]
    d_list = [1, 3]
    md = ModeledData(q, 300, rg_list=rg_list, d_list=d_list)
    if False:
        dplt.push()
        md.plot_components()
        dplt.pop()
    M = md.get_data()

    rank = len(rg_list)

class MfCorresDemo(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.currently_dragging = False
        self.current_n = None
        self.polygon = None
        Dialog.__init__(self, parent, "MF Tutorial Demo", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(20,10))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_press)
        self.mpl_canvas.mpl_connect('button_release_event', self.on_release)
        self.mpl_canvas.mpl_connect('motion_notify_event', self.on_motion)

        self.q, self.U, self.s, self.VT, self.axes, points, self.objects = demo_func_impl(fig)
        self.sqrt_s1s2 = np.sqrt(self.s[0]*self.s[1])
        self.ax6 = ax6 = self.axes[1][2]
        plot_text(ax6, "$X$", fontsize=FONTSIZE)
        xmin, xmax = ax6.get_xlim()
        ymin, ymax = ax6.get_ylim()
        tx = (xmin + xmax)/2
        ty = ymin*0.8 + ymax*0.2
        ax6.text(tx, ty, "Left  Click: any P, C", alpha=0.1, ha='center', fontsize=20)
        ty = ymin*0.9 + ymax*0.1
        ax6.text(tx, ty, "Right Click: povitive P, C only", alpha=0.1, ha='center', fontsize=20)

        self.points = points
        self.draw_points()
        self.mpl_canvas.draw()

    def draw_points(self):
        points = self.points
        points_ = np.array(points)
        #  print(points)
        # print(points_)
        for point, pobj in zip(points, self.objects):
            p, = pobj
            p.set_xdata([point[0]])
            p.set_ydata([point[1]])

        self.px = points_[:,0]
        self.py = points_[:,1]
        vx = np.sum(self.px)
        vy = np.sum(self.py)
        poly_points = [(0, 0), points[0], (vx,vy), points[1]]
        if self.polygon is None:
            self.polygon = Polygon(poly_points, alpha=0.1)
        else:
            self.polygon.set_xy(poly_points)

        self.ax6.add_patch(self.polygon)

    def on_press(self, event):
        if event.dblclick or True:
            self.draw_next_rand_point(event.button)
            return

        if event.inaxes != self.ax6:
            return

        print('on_press')
        self.currently_dragging = True
        dist = (self.px - event.xdata)**2 + (self.py - event.ydata)**2
        n = np.argmin(dist)
        print('n=', n)
        self.current_n = n

    def on_release(self, event):
        if not self.currently_dragging:
            return

        print('on_release')
        self.update_point(event)
        self.currently_dragging = False
        self.current_n = None
        self.redraw_decomp()
        self.mpl_canvas.draw()

    def on_motion(self, event):
        if not self.currently_dragging:
            return

        print('on_motion')
        n = self.current_n
        p, = self.objects[n]
        p.set_xdata([event.xdata])
        p.set_ydata([event.ydata])
        self.mpl_canvas.draw()

    def update_point(self, event):
        n = self.current_n
        self.points[n] = (event.xdata, event.ydata)
        self.draw_points()

    def redraw_decomp(self):
        ax8, ax5, _, ax7 = self.axes[1]
        for ax in [ax5, ax7, ax8]:
            ax.cla()
        draw_PQ([ax5, ax7, ax8], self.q, self.points, self.U, self.s, self.VT)

    def draw_next_rand_point(self, button):
        print('draw_next_rand_point')
        n = 0
        while True:
            # print([n])
            points = generate_random_points(self.s)
            # points = generate_random_points()
            if button == 1 or self.positive_PC(points):
                break
            n += 1
            if n > 1000000:
                break
        print([n])
        self.points = points
        self.draw_points()
        self.redraw_decomp()
        self.mpl_canvas.draw()

    def positive_PC(self, points):
        rank = len(points)
        a, b = points[0]
        c, d = points[1]
        s1, s2 = self.s[0:2]
        e, f, g, h = solve_efgh(a, b, c, d, s1, s2)
        X = np.array([[a, c],[b, d]])
        YT = np.array([[e, f],[g, h]])
        # print(np.dot(X, YT))
        # print(s1, s2)
        P = np.dot(self.U[:, 0:rank], X)
        C = np.dot(YT, self.VT[0:rank,:])
        return np.all(P >= 0) and np.all(C >= 0)



"""
    Optimizer.StrategicOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from scipy.optimize import basinhopping
LOG_IDENTIFICATION = "Strategic Optimizer"

class StrategicOptimizer:
    def __init__(self, optimizer, indeces):
        self.logger = logging.getLogger(__name__)
        self.optimizer = optimizer
        self.indeces = indeces
        self.logger.info("%s constructed with indeces=%s", LOG_IDENTIFICATION, str(indeces))

    def solve(self, init_params, real_bounds=None, niter=100, seed=None, open_mode="w", debug=False):
        self.logger.info("strategic solve started with init_params=%s", str(init_params))
        self.optimizer.prepare_for_optimization(init_params, real_bounds=real_bounds)

        self.cb_fh = open("callback.txt", open_mode)
        self.optimizer.cb_fh = self.cb_fh
        self.optimizer.write_init_callback_txt(init_params)

        self.temp_params = self.optimizer.to_norm_params(init_params)

        bounds = [(0, 10)]*len(self.indeces)
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)
        result = basinhopping(self.objective_func_wrapper, self.temp_params[self.indeces],
                              callback=self.minima_callback,
                              niter=niter,
                              seed=seed,
                              minimizer_kwargs=minimizer_kwargs)

        self.cb_fh.close()
        self.temp_params[self.indeces] = result.x
        result.x = self.temp_params
        return result
    
    def objective_func_wrapper(self, p):
        self.temp_params[self.indeces] = p
        real_params = self.optimizer.to_real_params(self.temp_params)
        return self.optimizer.objective_func(real_params)
    
    def minima_callback(self, x, f, accept):
        self.temp_params[self.indeces] = x
        return self.optimizer.minima_callback(self.temp_params, f, accept)



"""

    Models.ModelUtilsException.py

    Copyright (c) 2024, SAXS Team, KEK-PF

"""
import molass_legacy.KekLib.DebugPlot as plt

def get_range_exceptionally(x, model, params, debug=False):
    print("get_range_exceptionally: params=", params)
    y = model(x, params)
    h, m, s, t_ = params[0:4]
    f = max(x[0], m-2*s)
    t = min(x[-1], m+2*s)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("get_range_exceptionally")
            ax.plot(x, y)
            ax.axvspan(f, t, alpha=0.5)
            plt.show()

    return f, t


"""

    ZeroExtrapolator.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""
import time
import numpy                as np
import logging
from scipy                  import stats
# from lmfit                  import minimize, Parameters
from molass_legacy.KekLib.LmfitThreadSafe import minimize, Parameters
from molass_legacy.KekLib.BasicUtils import ordinal_str
import molass_legacy.KekLib.OurStatsModels as sm
from molass_legacy.AutorgKek.DataModels import GuinierPorodLmfit
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.KekLib.ProgressInfo import on_stop_raise
# from molass_legacy.LRF.LrfInfo import LrfInfo

NUM_EXTRAPOLATION_POINTS    = 5
USE_ERROR_PROPAGATION   = True
USE_MOORE_PENROSE_ONLY  = False
USE_LRF_RESULT_POOL = True
MIN_G_LN    = -100
MIN_G       = np.exp( MIN_G_LN )
IE_MIN_WIDTH        = 30
IE_B_MAX_MAX        = 0.1
IE_B_MIN_MAX        = 0.01
IE_B_ROUGH_BOUNDARY = 0.1
IE_P_VALUE_BOUNDARY = 0.5
NEXT_TEST_SPAN      = 5
RESET_SPAN          = 50
DYNAMIC_TEST_OK_COUNT   = 3
SEE_AQ_DEVIATION    = False
ACCEPTABLE_AQ_DEVIATION = 0.01
TO_RECOMPUTE_BQ_DEVIATION = 0.0001

def is_boundary_reg_params( reg_params ):
    # return False
    if reg_params is None:
        return False    # meaning not a cadidate

    r_value, p_value, std_err = reg_params
    return r_value > -0.95 or p_value > IE_P_VALUE_BOUNDARY or std_err > 5

def is_boundary_candidate( max_y, y, slope, p_value, j=None, log_aq_dev=None ):
    max_y_ = max( IE_B_MIN_MAX, min( IE_B_MAX_MAX, max_y ) )
    y_average = abs( np.average( y[-IE_MIN_WIDTH:] ) ) / max_y_

    if False and ( j is None or j > 100 and j < 300 ):
         print( [j], 'y_average=', y_average , 'slope=', slope, 'p_value=', p_value )

    if log_aq_dev is not None and log_aq_dev > ACCEPTABLE_AQ_DEVIATION:
        return True

    if y_average < IE_B_ROUGH_BOUNDARY:
        return p_value > IE_P_VALUE_BOUNDARY
    else:
        return False

def is_boundary_candidate_simple( y, slope, p_value ):
    return (    abs( np.average( y[-20:] ) ) < IE_B_ROUGH_BOUNDARY
            and abs( slope ) < IE_SLOPE_BOUNDARY
            and p_value > IE_P_VALUE_BOUNDARY )

def make_cx_vector_impl( use_elution_models, conc_factor, ad, paired_range, indeces, c_vector, jvector ):
    if use_elution_models:
        """
        temporary fix using jvector to avoid emg rare case trouble
        """
        wc_vector = np.zeros(len(jvector))
        for elm_rec in paired_range.elm_recs:
            e   = elm_rec[0]
            fnc = elm_rec[1]
            wc_vector += fnc(jvector)
        wc_vector *= conc_factor
        c_vector = wc_vector[indeces]

    zero_c  = 0.0
    min_c   = np.min( c_vector )
    max_c   = np.max( c_vector )
    min_c_1 = min_c * 4/5

    if ad == 0:
        cx_vector = np.hstack( [ np.linspace( zero_c, min_c_1, NUM_EXTRAPOLATION_POINTS ), c_vector ] )
    else:
        cx_vector = np.hstack( [ c_vector, np.linspace( min_c_1, zero_c, NUM_EXTRAPOLATION_POINTS ) ] )

    return cx_vector, min_c, max_c

class ZeroExtrapolator:
    def __init__( self, qvector, preview_params, serial_data, mapped_info, applied_ranges, known_info_list=None ):
        self.logger     = logging.getLogger( __name__ )
        self.qvector    = qvector
        self.sd         = serial_data
        self.mapped_info    = mapped_info
        self.applied_ranges = applied_ranges
        self.known_info_list = known_info_list
        self.conc_factor    = serial_data.conc_factor
        self.logger.info("lrf executer has been constructed with conc_factor=%g.", self.conc_factor)
        self.zx_boundary        = get_setting( 'zx_boundary' )
        self.num_q_points       = get_setting( 'zx_num_q_points' )
        self.zx_build_method    = get_setting( 'zx_build_method' )
        self.zx_boundary_method = get_setting( 'zx_boundary_method' )
        self.ignore_all_bqs     = get_setting('ignore_all_bqs')
        self.ignore_bq_list     = get_setting('ignore_bq_list')
        self.recompute_regboundary  = get_dev_setting( 'recompute_regboundary' )

        pdata, popts = preview_params
        self.use_elution_models = popts.use_elution_models

        if USE_LRF_RESULT_POOL:
            from molass_legacy.LRF.LrfResultPool import LrfResultPool
            self.pool = LrfResultPool(pdata, popts)
            self.pool.run_solver()
            self.logger.info("ZeroExtrapolator: using LRF Ppol.")
        else:
            assert False

        self.suppress_defer_test    = get_dev_setting( 'suppress_defer_test' )
        self.zx_add_constant        = get_dev_setting( 'zx_add_constant' )
        if self.zx_add_constant == 1:
            self.logger.info( 'Constant term will be added to the regression formulation.' )

    def make_cx_vector( self, ad, paired_range, indeces, c_vector ):
        return make_cx_vector_impl(self.use_elution_models, self.conc_factor, ad, paired_range, indeces, c_vector, self.sd.jvector)

    def extrapolate( self, m, ad, range_no, intensities, c_vector, cx_vector, max_c, boundary_q=0.01, guinier_boundary=None, temp_folder=None, debug_plot=False, debug_save=False ):
        self.stop_check()

        ones = np.ones( len(cx_vector) )
        # I(s) = C * A(s) + C**2 * B(s)
        # I(s)/C = A(s) + C * B(s)
        cx_matrix = np.array( [ ones, cx_vector ] ).T

        # self.logger.info( 'debug: guinier_boundary=' + str(guinier_boundary) )
        self.guinier_boundary   = guinier_boundary
        self.add_reg_candidate  = True

        result_with_reg, lrf_info = self.extrapolate_impl_MP_inverse( m, ad, range_no, intensities, c_vector, cx_vector, max_c, cx_matrix, boundary_q )

        return result_with_reg[0:-1], lrf_info

    def do_debug_save( self, m, ad, result, temp_folder ):
        if self.zx_boundary_method != 'AUTO' or self.guinier_boundary is None or temp_folder is None:
            return

        from molass_legacy.KekLib.NumpyUtils import np_savetxt

        aq_latest_list  = result[-1]
        aq_latest_array = np.array( aq_latest_list )
        start = self.guinier_boundary+1
        slice_ = slice(start,start+len(aq_latest_array))
        x_ = self.qvector[slice_]
        aq_latest_file = temp_folder + '/aq_latest_array-%d-%d.csv' % (m, ad)
        print( x_.shape, aq_latest_array.shape )
        np_savetxt( aq_latest_file, np.vstack( [ x_, aq_latest_array.T ] ).T )

    def do_debug_plot( self, results ):
        if self.zx_boundary_method != 'AUTO' or self.guinier_boundary is None:
            return

        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(15,8))

        max_lim_list = []

        param_types = [ '(no reg)', '(with reg)' ]

        for i, result in enumerate( results ):
            aq_latest_list  = result[-1]
            boundary_j      = result[-2]

            aq_latest_array = np.array( aq_latest_list )
            start = self.guinier_boundary+1
            slice_ = slice(start,start+len(aq_latest_array))
            x_ = self.qvector[slice_]

            for k, item in enumerate(['r_value', 'p_value', 'stderr']):
                ax = axes[i, k]
                ax.set_title( param_types[i] + ' ln(A(q)) 30p-segment ' + item )
                y_ = aq_latest_array[:,k]
                ax.plot( x_, y_ )
                if boundary_j is not None:
                    j_ = boundary_j - start
                    if j_ >= 0 and j_ < len(x_):
                        ax.plot( x_[j_], y_[j_], 'or' )

                if i == 0:
                    max_lim_list.append( [ ax.get_xlim(), ax.get_ylim() ] )
                else:
                    ax.set_xlim( max_lim_list[k][0] )
                    ax.set_ylim( max_lim_list[k][1] )

        plt.tight_layout()
        plt.show()

    def extrapolate_impl_MP_inverse( self, m, ad, range_no, intensities, c_vector, cx_vector, max_c, cx_matrix, boundary_q  ):
        self.boudary_test_deferred  = False
        self.boundary_j = None      # to be removed
        self.bq_max = 0             # to be removed

        ones = np.ones( len(cx_vector) )
        # I(s) = C * A(s) + C**2 * B(s)
        # I(s)/C = A(s) + C * B(s)
        cx_matrix = np.array( [ ones, cx_vector ] ).T

        data_list = []
        error_list = []
        for intensity in intensities:
            data_list.append( intensity.orig_array[:,1] )
            error_list.append( intensity.orig_array[:,2] )

        self.aq_max_c = data_list[-1 if ad == 0 else 0] / max_c
        self.log_aq_span = abs( np.log( np.average( self.aq_max_c[10:20]/self.aq_max_c[-20:-10] ) ) )

        data_matrix = np.vstack( data_list )
        error_matrix = np.vstack( error_list )
        print( 'data_matrix.shape=',data_matrix.shape )

        self.stop_check()

        C   = c_vector
        C2  = C*C
        C_  = np.array( [C, C2] ).T
        Cpinv = np.linalg.pinv( C_ )        # Moore-Penrose inverse

        if USE_MOORE_PENROSE_ONLY:
            P   = np.dot( Cpinv, data_matrix )
        else:
            if USE_LRF_RESULT_POOL:
                A, B, Z, E, lrf_info, C = self.pool.solver_results[range_no]
                P = np.vstack( [A, B] )
            else:
                assert False

        """
        
        """
        ze_array = np.dot( cx_matrix, P ).T
        param_array = P.T
        # ab_error_array = np.sqrt( np.dot( Cpinv**2, error_matrix**2 ) ).T
        ab_error_array = np.array(E[0:2]).T

        # errors for I(s) = C * A(s) + C**2 * B(s)
        ze_error_array = np.sqrt( np.dot( ab_error_array**2, cx_matrix.T**2, ) )

        if False:
            self.logger.info("ze_array.shape=%s, param_array.shape=%s, ab_err___array.shape=%s, ze_err___array.shape=%s",
                            str(ze_array.shape), str(param_array.shape), str(ab_error_array.shape), str(ze_error_array.shape))

        aq_latest_list = []
        return [data_matrix.T/c_vector, ze_array, param_array, ab_error_array, ze_error_array, self.boundary_j, aq_latest_list], lrf_info

    def dump_data(self, m, ad, C):
        import os
        from molass_legacy.KekLib.NumpyUtils import np_savetxt
        analysis_folder = get_setting('analysis_folder')
        file = os.path.join(analysis_folder, "C-%d-%d.dat" % (m, ad))
        np_savetxt(file, C.T)

    def stop_check( self ):
        on_stop_raise()

class GuinierPorodAnalyzer:
    def __init__( self, Q, I ):
        self.x  = Q**2
        self.y  = np.log(I)
        self.y[ I <= 0 ] = MIN_G_LN
        self.Q  = Q
        self.I  = I
        """
        I_positive = I > 0
        Q_ = Q[I_positive]
        I_ = I[I_positive]
        self.x  = Q_**2
        self.y  = np.log(I_)
        self.Q  = Q_
        self.I  = I_
        """

    def fit( self ):
        slope, intercept, r_value, p_value, std_err = stats.linregress( self.x, self.y )
        Rg  = np.sqrt( 3*(-slope) ) if slope < 0 else None
        G   = max( MIN_G, np.exp( intercept ) )
        d   = 2.5

        guinier_porod = GuinierPorodLmfit()

        params = Parameters()
        params.add('G',     value= G,  min=0 )
        params.add('Rg',    value= Rg, min=1.0, max=500.0 )
        params.add('d',     value= d,  min=1,   max=4 )

        result = minimize( guinier_porod, params, args=( self.Q, self.I, 1 ) )
        G_, Rg_, d_ = result.params['G'].value, result.params['Rg'].value, result.params['d'].value
        G_stdev = 0
        Rg_stdev = 0

        return G_, G_stdev, Rg_, Rg_stdev



"""
    KekLib.OnTheFlyUtils.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF
"""
import threading
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.DebugPlot import push, pop, get_parent

class ProgressDialog(Dialog):
    def __init__(self, parent, progress_queue, title="Progress", length=400, num_steps=10, interval=500):
        self.progress_queue = progress_queue
        self.length = length
        self.num_steps = num_steps
        self.interval = interval
        Dialog.__init__(self, parent, title, visible=False)

    def show(self):
        self._show()

    def body(self, body_frame): 
        self.mpb = ttk.Progressbar(body_frame,orient ="horizontal", length=self.length, mode="determinate")
        self.mpb.pack(padx=5, pady=5)
        self.mpb["maximum"] = self.num_steps
        self.update_progress()

    def update_progress(self):
        step = self.get_progress()
        if step is None:
            self.after(self.interval, self.update_progress)
            return

        if step < 0:
            self.cancel()

        if step < self.num_steps:
            self.mpb["value"] = step
            self.after(self.interval, self.update_progress)
            self.update()
        else:
            self.mpb["value"] = self.num_steps
            self.update()
            self.cancel()            

    def get_progress(self):
        return self.progress_queue.get()

def show_progress(func, progress_queue, num_steps, parent=None):
    thread = threading.Thread(target=func)
    thread.start()
    push()
    if parent is None:
        parent = get_parent()
    progress_queue.put(0)
    dialog = ProgressDialog(parent, progress_queue, num_steps=num_steps)
    dialog.show()
    pop()
    thread.join()


# coding: utf-8
"""
    LinearEquations.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.ticker as ticker
from OurManim import latex_init
import molass_legacy.KekLib.DebugPlot as plt

def show_figs():
    latex_init()

    fig, axes = plt.subplots(ncols=3, figsize=(21,7))
    a, b, c = 2, 1, 3

    fontsize = 30
    ax1, ax2, ax3 = axes
    ax1.set_title("Under-determined System", fontsize=fontsize)
    ax2.set_title("Exactly-determined System", fontsize=fontsize)
    ax3.set_title("Over-determined System", fontsize=fontsize)

    x = np.linspace(-0.5, 3, 10)
    y = a - x
    for ax in axes:
        ax.plot(x, y, label='$x+y=%g$' % a)
    for ax in axes[1:]:
        ax.plot(x, x-b, label='$x-y=%g$' % b)
    for ax in axes[2:]:
        ax.plot(x, (c-x)/2, label='$x+2y=%g$' % c)

    v = np.linalg.pinv(np.array([[1,1]]))@np.array([[a]])
    ax1.plot(*v, 'o', color='red', label=r'$ {\begin{bmatrix} 1 & 1 \end{bmatrix}}^+ \begin{bmatrix} %g \end{bmatrix} $' % a)


    v = np.linalg.pinv(np.array([[1,1], [1,-1]])) @ np.array([[a],[b]])
    ax2.plot(*v, 'o', color='red', label=r'$ {\begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}}^+ \begin{bmatrix} %g \\ %g \end{bmatrix} $' % (a, b))

    v = np.linalg.pinv(np.array([[1,1], [1,-1], [1,2]])) @ np.array([[a],[b],[c]])
    ax3.plot(*v, 'o', color='red', label=r'$ {\begin{bmatrix} 1 & 1 \\ 1 & -1 \\ 1 & 2 \end{bmatrix}}^+ \begin{bmatrix} %g \\ %g \\ %g \end{bmatrix} $' % (a, b, c))

    for ax in axes:
        ax.set_xlim(0,2.5)
        ax.set_ylim(0,2.5)
        ax.set_aspect('equal')
        ax.legend(loc="upper center", fontsize=20)
        ax.xaxis.set_major_locator(ticker.FixedLocator([0,1,2]))
        ax.yaxis.set_major_locator(ticker.FixedLocator([0,1,2]))

    fig.tight_layout()
    fig.subplots_adjust(top=0.9, bottom=0.1)
    plt.show()



"""
    ATSAS.ExecCheck.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import os
from .AutoRg import autorg_exe_array

def atsas_exec_check(parent, debug=False):
    if len(autorg_exe_array) == 0:
        return False

    exe_path = autorg_exe_array[0]

    ok = os.path.exists(exe_path)

    if not ok or debug:
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        folder = os.path.dirname(os.path.dirname(exe_path))
        MessageBox.showerror("ATSAS Installation Info Error",
            'Somethig is wrong with the ATSAS installation under\n'
            '"%s"\n'
            'Check the installation status at "Settings/Basic Settings"\n'
            'and try "Update" button if it is not the latest.'
            % folder,
            parent=parent
            )

        if debug:
            ok = False

    return ok



"""
    AutoRestrictor.py

    Copyright (c) 2020-2024, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
from scipy.stats import linregress
from scipy.interpolate import LSQUnivariateSpline
from molass_legacy.Elution.CurveUtils import simple_plot
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.Models.ElutionCurveModels import EGH
from molass_legacy.QuickAnalysis.ParammedModel import ParammedModel
from .MatchingPeaks import MatchingPeaks
from .TrimmingInfo import TrimmingInfo

SIGMA_POINT_RATIO = 10
MINOR_SIGMA_RATIO = 5
RECON_SIGMA_RATIO = 7
RECON_MINOR_SIGMA_RATIO = 3
EXTEND_LIMIT = 0.05
ACCEPTABLE_Y_RATIO = 0.05
N_SIGMAS_UNRELIABLE = 2.5   # chosen to modify so that 20191006_proteins5 be ok
MAX_SIGMA_WIDTH_RATIO = 0.7

def get_exact_emg_peaks_list(uv_curve, xr_curve, debug=False):
    # set_setting("local_debug", True)
    ret_uv_peaks = uv_curve.get_emg_peaks(debug=debug)
    # set_setting("local_debug", False)
    ret_xr_peaks = xr_curve.get_emg_peaks(debug=debug)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_exact_emg_peaks_list")

            simple_plot(ax1, uv_curve, color="blue", legend=False)
            x = uv_curve.x
            for epeak in ret_uv_peaks:
                y = epeak.get_model_y(x)
                ax1.plot(x, y, ":")

            simple_plot(ax2, xr_curve, color="orange", legend=False)
            x = xr_curve.x
            for epeak in ret_xr_peaks:
                y = epeak.get_model_y(x)
                ax2.plot(x, y, ":")

            for ax in ax1, ax2:
                ax.legend()
            fig.tight_layout()
            plt.show()

    return ret_uv_peaks, ret_xr_peaks

class PeakMathcingError(Exception):
    pass


class AutoRestrictor:
    def __init__(self, sd, cs):
        self.logger = logging.getLogger()
        self.sd = sd
        self.cs = cs
        self.extended_flags = np.zeros((2, 2), dtype=int)
        self.mpeaks = None
        """
                        0 : left,  1 : right
            0 : UV
            1 : Xray
        """

    def get_better_restriction(self, uv_restrict_list=None, xr_restrict_list=None, debug=False, figfile=None):
        try:
            return self.get_better_restriction_impl(uv_restrict_list, xr_restrict_list, debug=debug, figfile=figfile)
        except:
            # as in 20181203 with PEAK_INFO_FIND_RATIO=0.3
            log_exception(self.logger, "get_better_restriction failed: ")

            if debug:
                from .TrimmingDebugUtils import trimming_debug_plot
                trimming_debug_plot("get_better_restriction_impl except debug", self.sd, uv_restrict_list, xr_restrict_list)

            return uv_restrict_list, xr_restrict_list

    def get_better_restriction_impl(self, uv_restrict_list=None, xr_restrict_list=None, debug=False, figfile=None):
        """
        TODO:
            unify *_a_restrict setting in this method
            from molass_legacy.Trimming.PreliminaryRecognition.set_restrict_info
        """

        if uv_restrict_list is None:
            uv_e_restrict, uv_a_restrict = None, None
        else:
            uv_e_restrict, uv_a_restrict = uv_restrict_list
        if xr_restrict_list is None:
            xr_e_restrict, xr_a_restrict = None, None
        else:
            xr_e_restrict, xr_a_restrict = xr_restrict_list

        uv_wl_lower_bound = get_setting("uv_wl_lower_bound")
        assert uv_wl_lower_bound is not None
        wlvector = self.sd.lvector
        start = bisect_right(wlvector, uv_wl_lower_bound)
        if uv_a_restrict is None:
                size = len(wlvector)
                uv_a_restrict = TrimmingInfo(1, start, size, size)
        else:
            self.logger.info("taling max(%d, %d)", uv_a_restrict.start, start)
            uv_a_restrict.start = max(uv_a_restrict.start, start)

        self.logger.info("uv_wl_lower_bound=%g, uv_a_restrict=%s", uv_wl_lower_bound, str(uv_a_restrict))
        sd = self.sd
        cs = self.cs
        # uv_curve = sd.get_uv_curve()
        # xr_curve = sd.get_xray_curve()
        uv_curve = cs.a_curve
        xr_curve = cs.x_curve

        try:
            # uv_peaks, xr_peaks = self.get_major_emg_peaks_list(uv_curve, xr_curve, debug=debug)
            uv_peaks, xr_peaks = get_exact_emg_peaks_list(uv_curve, xr_curve, debug=False)

            new_xr_restrict = self.get_sigma_restriction(xr_curve, xr_peaks, xr_e_restrict, debug=debug)
            mapped_points, maping_info = self.get_mapped_points(new_xr_restrict, xr_curve, xr_peaks, uv_curve, uv_peaks, debug=debug)
            new_uv_restrict = self.get_sigma_restriction(uv_curve, uv_peaks, uv_e_restrict, mapped_points=mapped_points, op_restrict=new_xr_restrict, debug=debug)
            new_xr_restrict_save = new_xr_restrict
            new_xr_restrict = self.map_back_to_xray(maping_info, new_uv_restrict, new_xr_restrict)

            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                def plot_range(ax, restrict, color):
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    for j in (restrict.start, restrict.stop):
                        ax.plot([j, j], [ymin, ymax], color=color)

                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("get_better_restriction_impl: debug")
                    simple_plot(ax1, uv_curve, color="blue")
                    simple_plot(ax2, xr_curve, color="orange")
                    plot_range(ax1, new_uv_restrict, "yellow")
                    plot_range(ax2, new_xr_restrict_save, "cyan")
                    plot_range(ax2, new_xr_restrict, "yellow")
                    fig.tight_layout()
                    plt.show()

            new_e_restricts = [new_uv_restrict, new_xr_restrict]
        except:
            log_exception(self.logger, "get_better_restriction_impl error: ")
            new_e_restricts = self.get_them_roughly((uv_curve, uv_e_restrict), (xr_curve, xr_e_restrict), debug=debug)

        ret_info = [[new_e_restricts[0], uv_a_restrict], [new_e_restricts[1], xr_a_restrict]]

        curves = uv_curve, xr_curve
        emg_peaks_list = uv_peaks, xr_peaks
        old_e_restricts = uv_e_restrict, xr_e_restrict
        self.result_info = sd, curves, emg_peaks_list, old_e_restricts, ret_info, SIGMA_POINT_RATIO, self.mpeaks

        if debug or figfile is not None:
            from molass_legacy.Trimming.TrimmingDebugUtils import trimming_result_plot
            trimming_result_plot(*self.result_info, figfile=figfile)

        if (self.extended_flags[0,:] != self.extended_flags[1,:]).any():
            ret_info = self.correct_inconsistent_extensions(ret_info)

        return ret_info

    def get_them_roughly(self, uv_info, xr_info, debug=False):
        from molass_legacy.Peaks.RobustPeaks import RobustPeaks
        from molass_legacy.Trimming import get_mapped_info, get_wider_info

        ret_e_restricts = []
        for curve, e_restrict in [uv_info, xr_info]:
            x = curve.x
            y = curve.y
            rp = RobustPeaks(x, y)
            sigma_min, sigma_max = rp.get_limits_roughly(SIGMA_POINT_RATIO)

            size = len(x)
            if e_restrict is None or not e_restrict.flag:
                xmin, xmax = 0, size    # or size-1 ?
            else:
                xmin, xmax = e_restrict.start, e_restrict.stop

            ret_e_restricts.append(TrimmingInfo(1, max(xmin, sigma_min), min(xmax, sigma_max), size))

        A = self.cs.slope
        B = self.cs.intercept
        mapped_xr_info = get_mapped_info((A, B), ret_e_restricts[0].size, ret_e_restricts[1])
        ret_uv_info = get_wider_info(ret_e_restricts[0], mapped_xr_info)
        mapped_uv_info = get_mapped_info((1/A, -B/A), ret_e_restricts[1].size, ret_uv_info)
        ret_xr_info = get_wider_info(ret_e_restricts[1], mapped_uv_info)
        ret_e_restricts = [ret_uv_info, ret_xr_info]

        self.logger.info("elution trimming infos have been determined roughly due to a peak matching trouble.")

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, axes = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("get_them_roughly debug")

                def draw_trimming(ax, ylim, tinfo, color):
                    if tinfo is None:
                        return
                    for j in tinfo.start, tinfo.stop:
                        ax.plot([j, j], ylim, color=color)

                for ax, in_info, ret_info in zip(axes, [uv_info, xr_info], ret_e_restricts):
                    curve, e_restrict = in_info
                    simple_plot(ax, curve)
                    ylim = ax.get_ylim()
                    ax.set_ylim(ylim)
                    draw_trimming(ax, ylim, e_restrict, "cyan")
                    draw_trimming(ax, ylim, ret_info, "yellow")
                fig.tight_layout()
                plt.show()

        return ret_e_restricts

    def get_uv_sigma_points(self):
        return self.uv_sigma_info.start, self.uv_sigma_info.stop

    def get_major_emg_peaks_list(self, uv_curve, xr_curve, debug=False):
        self.mpeaks = MatchingPeaks(self.cs, uv_curve, xr_curve, debug=debug)
        uv_emg_peaks, xr_emg_peaks = self.mpeaks.get_matching_peak_lists()

        uv_acc_y = uv_curve.max_y * ACCEPTABLE_Y_RATIO
        xr_acc_y = xr_curve.max_y * ACCEPTABLE_Y_RATIO

        ret_uv_peaks = []
        ret_xr_peaks = []
        for uv_peak, xr_peak in zip(uv_emg_peaks, xr_emg_peaks):
            if uv_peak.top_y >= uv_acc_y or xr_peak.top_y >= xr_acc_y:
                ret_uv_peaks.append(uv_peak)
                ret_xr_peaks.append(xr_peak)

        return ret_uv_peaks, ret_xr_peaks

    def get_mapped_points(self, restrict_info, xr_curve, xr_peaks, uv_curve, uv_peaks, debug=False):
        slope = self.cs.slope
        intercept = self.cs.intercept

        def mapped_pos(k, p):
            if p is None:
                i = 0 if k == 0 else -1
                p = xr_curve.x[i]
            return int(round(p*slope + intercept))

        mapped_points = [mapped_pos(k, p) for k, p in enumerate([restrict_info.start, restrict_info.stop])]
        return mapped_points, (slope, intercept)

    def get_sigma_restriction(self, curve, emg_peaks, e_restrict,
            mapped_points=None,
            op_restrict=None,       # not used currently
            debug=False):

        x = curve.x
        size = len(x)

        num_peaks = len(emg_peaks)
        if num_peaks == 1:
            sigma_min, sigma_max = self.get_simple_sigma_points(curve, emg_peaks[0], debug=debug)
            info_L = (emg_peaks[0], sigma_min)
            info_R = (emg_peaks[0], sigma_max)
        else:
            sigma_min = self.get_simple_sigma_points(curve, emg_peaks[0], debug=debug)[0]
            sigma_max = self.get_simple_sigma_points(curve, emg_peaks[-1], debug=debug)[-1]
            info_L = (emg_peaks[0], sigma_min)
            info_R = (emg_peaks[-1], sigma_max)

        if mapped_points is not None:
            s_min, s_max = mapped_points
            sigma_min = min(sigma_min, s_min)
            sigma_max = max(sigma_max, s_max)

            if not self.cs.reliable:
                # consider to do this always
                sigma_min, sigma_max = self.modify_limits_using_moments(curve, sigma_min, sigma_max)

        sigma_info = Struct(jmin=sigma_min, jmax=sigma_max)

        if e_restrict is None or not e_restrict.flag:
            xmin, xmax = 0, size    # or size-1 ?
        else:
            xmin, xmax = e_restrict.start, e_restrict.stop

        trim_min = max(xmin, sigma_min)
        trim_max = min(xmax, sigma_max)

        try:
            recon_info = self.reconsider_using_lower_peak_model(curve, emg_peaks, trim_min, trim_max, debug=debug)
        except:
            log_exception(self.logger, "reconsider_using_lower_peak_model failed: ")
            recon_info = None

        if recon_info is None:
            final_min = trim_min
            final_max = trim_max
        else:
            final_min = max(xmin, min(sigma_min, recon_info.jmin))
            final_max = min(xmax, max(sigma_max, recon_info.jmax))

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_sigma_restriction debug")

                simple_plot(ax, curve, legend=False)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)

                for x in [final_min, final_max]:
                    ax.plot([x, x], [ymin, ymax], color="yellow", lw=2, label="final")

                if mapped_points is not None:
                    for x in mapped_points:
                        ax.plot([x, x], [ymin, ymax], ":", color="cyan", lw=2, label="mapped")

                ax.legend()
                fig.tight_layout()
                plt.show()

        extra = Struct(info_L=info_L, info_R=info_R, mapped_points=mapped_points, num_peaks=num_peaks,
                        sigma_info=sigma_info, recon_info=recon_info)

        return TrimmingInfo(1, final_min, final_max, size, extra=extra)

    def get_simple_sigma_points(self, curve, emgpeak, debug=False):
        return self.get_simple_sigma_points_impl(curve, emgpeak.opt_params, emgpeak.top_x, debug=debug)

    def get_simple_sigma_points_impl(self, curve, params, peak_top_x, force_ratio=None, debug=False):
        if debug:
            from molass_legacy.KekLib.DebugUtils import show_call_stack
            show_call_stack("----: ", indented_only=True)

        sigma = params[2]
        tau = params[3]

        if force_ratio is None:
            y = curve.spline(peak_top_x)
            if y > curve.max_y*0.1:
                ratio = SIGMA_POINT_RATIO
            else:
                ratio = MINOR_SIGMA_RATIO
        else:
            ratio = force_ratio

        # width = ratio*max(sigma, -tau)
        width = ratio*sigma     # tau < 0 is not valid according to stochastic theory
        f = max(curve.x[0], peak_top_x - width)
        width = ratio*max(sigma, tau)
        t = min(curve.x[-1], peak_top_x + width)

        width_ratio = width/len(curve.x)
        if width_ratio < MAX_SIGMA_WIDTH_RATIO:
            pass
        else:
            # as in SUB_TRN1
            raise AssertionError("width_ratio=%.3g >= MAX_SIGMA_WIDTH_RATIO" % width_ratio)

        ret_points = [int(round(v)) for v in (f, t)]
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Peaks.ElutionModels import egh
            print("sigma=", sigma, "tau=", tau, "ratio=", ratio)
            print("width=", width)
            x = curve.x
            h = params[0]
            mu = params[1]
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_simple_sigma_points_impl")
                simple_plot(ax, curve, legend=False)
                ax.plot(x, egh(x, h, mu, sigma, tau), ":", label="model")
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for j in ret_points:
                    ax.plot([j, j], [ymin, ymax], color="yellow")
                ax.legend()
                fig.tight_layout()
                plt.show()

        return ret_points

    def proof_plot(self, better_ret, pause=True, savefig=False, figfile=None):
        import molass_legacy.KekLib.DebugPlot as plt
        from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
        plt.push()
        # uv_curve = self.sd.get_uv_curve()
        # xr_curve = self.sd.get_xray_curve()
        uv_curve = self.cs.a_curve
        xr_curve = self.cs.x_curve
        uv_ret, xr_ret = better_ret
        uv_restrict = None if uv_ret is None else uv_ret[0]
        xr_restrict = None if xr_ret is None else xr_ret[0]

        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,6))
        ax1, ax2 = axes
        fig.suptitle("AutoRestrictor Proof for %s" % get_in_folder(), fontsize=16)

        ax1.plot(uv_curve.y)
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)
        if uv_restrict is not None:
            for x in [uv_restrict.start, uv_restrict.stop]:
                ax1.plot([x,x], [ymin, ymax], color='yellow')

        for rec in uv_curve.peak_info:
            x = rec[1]
            ax1.plot(x, uv_curve.spline(x), 'o', color='pink')

        for peak in uv_curve.get_emg_peaks():
            ax1.plot(peak.top_x, peak.top_y, 'o', color='red')

        ax2.plot(xr_curve.y)
        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        if xr_restrict is not None:
            for x in [xr_restrict.start, xr_restrict.stop]:
                ax2.plot([x,x], [ymin, ymax], color='yellow')

        for rec in xr_curve.peak_info:
            x = rec[1]
            ax2.plot(x, xr_curve.spline(x), 'o', color='pink')

        for peak in xr_curve.get_emg_peaks():
            ax2.plot(peak.top_x, peak.top_y, 'o', color='red')

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        if pause:
            plt.show()
        else:
            from time import sleep
            plt.show(block=False)
            sleep(0.5)
        if savefig:
            import os
            from molass_legacy._MOLASS.SerialSettings import get_setting
            temp_folder = get_setting('temp_folder')
            path = os.path.join(temp_folder, figfile)
            fig.savefig(path)
        plt.pop()

    def correct_inconsistent_extensions(self, ret_info):
        flags = self.extended_flags
        self.logger.info("correcting the inconsistent extensions detected: %s.", str(list(flags)))
        lr_id = 0
        for uv_flag, xr_flag in zip(flags[0,:], flags[1,:]):
            if uv_flag == xr_flag:
                pass
            else:
                uv_xr_id =  uv_flag
                self.logger.info("extend also %s %s.", ('Xray' if uv_xr_id else 'UV'), ('right' if lr_id else 'left'))
                ret_info[uv_xr_id] = self.set_correponding_point(uv_xr_id, lr_id, ret_info[uv_xr_id], ret_info[1 - uv_xr_id])
            lr_id += 1
        return ret_info

    def set_correponding_point(self, uv_xr_id, lr_id, info_to, info_from):
        if uv_xr_id:
            # correcting Xray from molass_legacy.UV
            # UV => Xray
            A = 1/self.cs.slope
            B = - self.cs.intercept/self.cs.slope
        else:
            # correcting UV from Xray
            # Xray => UV
            A = self.cs.slope
            B = self.cs.intercept

        """
        print('info_to=', info_to)
        print('info_from=', info_from)
        info_to= [Info(1, 0, 804, 805), None]
        info_from= [Info(1, 6, 401, 402), None]
        """

        k = lr_id + 1
        from_value = info_from[0][k]
        to_value = info_to[0][k]
        to_max = info_to[0][-1]
        set_value = max(0, min(to_max-1, int(A*from_value + B)))
        self.logger.info("correcting %d to %d by %d.", to_value, set_value, from_value)
        set_list_ = list(info_to[0])
        set_list_[k] = set_value
        info_to[0] = TrimmingInfo(*set_list_)
        return info_to

    def modify_limits_using_moments(self, curve, xmin, xmax):
        try:
            mean, variance = curve.compute_moments()
            assert variance > 0
            sigma = np.sqrt(variance)
            print('modify_limits_using_moments: mean=', mean, 'sigma=', sigma)
            ret_xmin = int(min(mean - N_SIGMAS_UNRELIABLE*sigma, xmin))
            ret_xmax = int(max(mean + N_SIGMAS_UNRELIABLE*sigma, xmax))
            if ret_xmin < xmin or ret_xmax > xmax:
                self.logger.info("modified sigma limits from %s to %s due to unreliable mapping.", str((xmin, xmax)), str((ret_xmin, ret_xmax)))
        except:
            ret_xmin = xmin
            ret_xmax = xmax
        return ret_xmin, ret_xmax

    def reconsider_using_lower_peak_model(self, curve, emg_peaks, trim_min, trim_max, debug=False):
        num_peaks = len(emg_peaks)
        if num_peaks > 1:
            # currently support only when num_peaks == 1
            return None

        main_emg_peak = emg_peaks[0]
        eslice = slice(trim_min, trim_max)
        x = curve.x[eslice]
        y = curve.y[eslice]
        low_part = y < curve.max_y * 0.25
        x_ = x[low_part]
        y_ = y[low_part]
        knots = np.linspace(x[1], x[-2], 10)
        spline = LSQUnivariateSpline(x_, y_, knots)
        sy = spline(x)

        model = EGH()
        params = model.guess(sy, x=x)
        out = model.fit(sy, params, x=x)
        fy = model.eval(out.params, x=x)

        peak_top_x = main_emg_peak.top_x

        orig_x = curve.x
        orig_y = curve.y

        curve_proxy = Struct(x=orig_x, max_y=curve.max_y, spline=spline)

        try:
            recon_min, recon_max = self.get_simple_sigma_points_impl(curve_proxy, out.params, peak_top_x,
                                                                        force_ratio=RECON_SIGMA_RATIO)
        except:
            log_exception(self.logger, "get_simple_sigma_points_impl: ")
            raise RuntimeError("get_simple_sigma_points_impl failure")

        prm_model = ParammedModel([(model, out.params)])
        recon_info = Struct(jmin=recon_min, jmax=recon_max, prm_model=prm_model)
        recon_info = self.try_adding_minor_peak(curve, main_emg_peak, recon_info)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            plt.push()
            fig, ax = plt.subplots()
            ax.set_title("reconsider_using_lower_peak_model debug")
            ax.plot(orig_x, orig_y)
            ax.plot(x_, y_)
            ax.plot(x, sy)
            ax.plot(x, fy)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for px in [recon_min, recon_max]:
                ax.plot([px, px], [ymin, ymax], ":", color="red")
            plt.show()
            plt.pop()

        return recon_info

    def try_adding_minor_peak(self, curve, main_emg_peak, recon_info):
        peak_top_x = main_emg_peak.top_x
        candidates = []
        for k, info in enumerate(curve.peak_info):
            ptx = info[1]
            if ptx < peak_top_x:
                candidates.append(info)

        if len(candidates) > 0:
            # select the tallest minor peak
            candidate = sorted(candidates, key=lambda info: -curve.y[info[1]])[0]
            eslice = slice(candidate[0], candidate[2])
            x_ = curve.x[eslice]
            y_ = curve.y[eslice]
            model = EGH()
            params = model.guess(y_, x=x_)
            out = model.fit(y_, params, x=x_)
            fy = model.eval(out.params, x=x_)

            jmin = max(0, int(round(candidate[1] - abs(out.params["tau"]) * RECON_MINOR_SIGMA_RATIO)))

            if False:
                import molass_legacy.KekLib.DebugPlot as plt
                from molass_legacy.Elution.CurveUtils import simple_plot
                plt.push()
                fig, ax = plt.subplots()
                ax.set_title("try_adding_minor_peak debug")
                simple_plot(ax, curve)
                ax.plot(x_, fy)
                ax.plot([jmin], model.eval(out.params, x=[jmin]), "o", color="yellow")
                plt.show()
                plt.pop()

            prm_model = recon_info.prm_model
            prm_model.append((model, out.params))
            ret_info = Struct(jmin=jmin, jmax=recon_info.jmax, prm_model=prm_model)
        else:
            ret_info = recon_info

        return ret_info

    def map_back_to_xray(self, maping_info, uv_restrict, xr_restrict):
        a, b = maping_info
        # j = a*i + b
        # i = (j - b)/a
        jmin, jmax = [ int(round((j - b)/a)) for j in [uv_restrict.start, uv_restrict.stop]]

        size = xr_restrict.size
        extra = xr_restrict.extra
        return TrimmingInfo(1, max(0, jmin), min(size, jmax), size, extra=extra)
